import unittest
import os
import shutil
import warnings
from ml4ir.base.data import ranklib_helper
import pandas as pd
import yaml
from ml4ir.base.features.feature_config import ExampleFeatureConfig, SequenceExampleFeatureConfig
from ml4ir.base.config.keys import TFRecordTypeKey
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.config.keys import DataFormatKey
from ml4ir.base.features.preprocessing import *
from ml4ir.base.io import file_io, local_io

warnings.filterwarnings("ignore")

INPUT_DIR = "ml4ir/applications/ranking/tests/data/ranklib/"
KEEP_ADDITIONAL_INFO = 0
NON_ZERO_FEATURES_ONLY = 0


class TestRanklibConversion(unittest.TestCase):
    def setUp(self):
        self.feature_config_yaml = INPUT_DIR + 'feature_config.yaml'
        self.feature_config_yaml_convert_to_clicks = INPUT_DIR + \
            'feature_config_convert_to_clicks.yaml'

    def tearDown(self):
        if os.path.exists(os.path.join(INPUT_DIR, "tfrecord")):
            shutil.rmtree(os.path.join(INPUT_DIR, "tfrecord"))

    def parse_config(self, tfrecord_type: str, feature_config, io) -> SequenceExampleFeatureConfig:
        if feature_config.endswith(".yaml"):
            feature_config = io.read_yaml(feature_config)
        else:
            feature_config = yaml.safe_load(feature_config)

        return SequenceExampleFeatureConfig(feature_config, None)

    def test_ranklib_in_ml4ir(self):
        """Creates a relevance dataset using ranklib format. Labels are graded relevance"""

        io = local_io.LocalIO()
        exFeatureConfig = self.parse_config(
            TFRecordTypeKey.SEQUENCE_EXAMPLE, self.feature_config_yaml, io)
        preprocessing_keys_to_fns = {}
        if 'preprocessing_info' in exFeatureConfig.get_label():
            if exFeatureConfig.get_label()['preprocessing_info'][0]['fn'] == 'convert_label_to_clicks':
                preprocessing_keys_to_fns['convert_label_to_clicks'] = convert_label_to_clicks

        dataset = RelevanceDataset(
            data_dir=INPUT_DIR,
            data_format=DataFormatKey.RANKLIB,
            feature_config=exFeatureConfig,
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            batch_size=1,
            file_io=io,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            logger=None,
            keep_additional_info=KEEP_ADDITIONAL_INFO,
            non_zero_features_only=NON_ZERO_FEATURES_ONLY,
            max_sequence_size=319,
        )
        non_one_hot = False
        chk = [e for e in dataset.train]
        for e in chk:
            if sum(e[1][0]).numpy() > 1:
                non_one_hot = True
                break
        assert non_one_hot == True
        assert len(chk) == 49

        non_one_hot = False
        chk = [e for e in dataset.validation]
        for e in chk:
            if sum(e[1][0]).numpy() > 1:
                non_one_hot = True
                break
        assert non_one_hot == True
        assert len(chk) == 49

        non_one_hot = False
        chk = [e for e in dataset.test]
        for e in chk:
            if sum(e[1][0]).numpy() > 1:
                non_one_hot = True
                break
        assert non_one_hot == True
        assert len(chk) == 49

    def test_ranklib_in_ml4ir_click_conversion(self):
        """Creates a relevance dataset using ranklib format. Labels are converted to clicks graded relevance"""
        io = local_io.LocalIO()
        exFeatureConfig = self.parse_config(TFRecordTypeKey.SEQUENCE_EXAMPLE,
                                            self.feature_config_yaml_convert_to_clicks, io)
        preprocessing_keys_to_fns = {}
        if exFeatureConfig.get_label()['preprocessing_info'][0]['fn'] == 'convert_label_to_clicks':
            preprocessing_keys_to_fns['convert_label_to_clicks'] = convert_label_to_clicks

        dataset = RelevanceDataset(
            data_dir=INPUT_DIR,
            data_format=DataFormatKey.RANKLIB,
            feature_config=exFeatureConfig,
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            batch_size=1,
            file_io=io,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            logger=None,
            keep_additional_info=KEEP_ADDITIONAL_INFO,
            non_zero_features_only=NON_ZERO_FEATURES_ONLY,
            max_sequence_size=319,
        )
        chk = [e for e in dataset.train]
        for e in chk:
            assert max(e[1][0]).numpy() == 1
        assert len(chk) == 49

        chk = [e for e in dataset.validation]
        for e in chk:
            assert max(e[1][0]).numpy() == 1
        assert len(chk) == 49

        chk = [e for e in dataset.test]
        for e in chk:
            assert max(e[1][0]).numpy() == 1
        assert len(chk) == 49


if __name__ == "__main__":
    unittest.main()



import os
import numpy as np
import tensorflow as tf

from ml4ir.applications.ranking.tests.test_base import RankingTestBase
from ml4ir.applications.ranking.model.losses import pointwise_losses
from ml4ir.applications.ranking.model.losses import listwise_losses


class RankingModelTest(RankingTestBase):

    def setUp(self):

        super().setUp()

        self.logits = tf.constant(
            [[0.19686124, 0.73846658, 0.17136828, -0.11564828, -0.3011037],
             [-1.47852199, -0.71984421, -0.46063877, 1.05712223, 0.34361829],
             [-1.76304016, 0.32408397, -0.38508228, -0.676922, 0.61167629]])
        self.y_true = tf.constant(
            [[0., 0., 1., 0., 0.],
             [0., 1., 0., 0., 0.],
             [0., 1., 0., 0., 0.]])
        self.mask = tf.constant(
            [[1., 1., 1., 1., 1.],
             [1., 1., 1., 1., 0.],
             [1., 1., 0., 0., 0.]])

    def test_sigmoid_cross_entropy(self):
        """Test the sigmoid cross entropy pointiwse loss object"""
        loss = pointwise_losses.SigmoidCrossEntropy()
        activation_op = loss.get_final_activation_op(output_name="y_pred")
        loss_fn = loss.get_loss_fn(mask=self.mask)

        y_pred = activation_op(logits=self.logits, mask=self.mask)
        assert np.isclose(y_pred[0][0].numpy(), 0.54905695, atol=1e-5)
        assert np.isclose(y_pred[2][4].numpy(), 0.64832306, atol=1e-5)

        assert np.isclose(loss_fn(self.y_true, y_pred), 0.6905699, atol=1e-5)

    def test_softmax_cross_entropy(self):
        """Test the softmax cross entropy listwise loss object"""
        loss = listwise_losses.SoftmaxCrossEntropy()
        activation_op = loss.get_final_activation_op(output_name="y_pred")
        loss_fn = loss.get_loss_fn(mask=self.mask)

        y_pred = activation_op(logits=self.logits, mask=self.mask)

        assert np.isclose(y_pred[0][0].numpy(), 0.19868991, atol=1e-5)
        assert np.isclose(y_pred[2][4].numpy(), 0., atol=1e-5)

        assert np.isclose(loss_fn(self.y_true, y_pred), 1.306335, atol=1e-5)

    def test_rank_one_list_net(self):
        """Test the rank-one listnet listwise loss object"""
        loss = listwise_losses.RankOneListNet()
        activation_op = loss.get_final_activation_op(output_name="y_pred")
        loss_fn = loss.get_loss_fn(mask=self.mask)

        y_pred = activation_op(logits=self.logits, mask=self.mask)

        assert np.isclose(y_pred[0][0].numpy(), 0.19868991, atol=1e-5)
        assert np.isclose(y_pred[2][4].numpy(), 0., atol=1e-5)

        assert np.isclose(loss_fn(self.y_true, y_pred), 2.1073625, atol=1e-5)



import string
import re
import tensorflow as tf

from ml4ir.base.features.feature_fns.categorical import categorical_indicator_with_vocabulary_file
from ml4ir.base.io.file_io import FileIO


class PreprocessingMap:
    """
    Class defining a map of predefined and custom preprocessing functions
    """

    def __init__(self):
        """
        Instantiate a PreprocessingMap object with predefined functions
        """
        self.key_to_fn = {
            preprocess_text.__name__: preprocess_text,
            split_and_pad_string.__name__: split_and_pad_string,
            natural_log.__name__: natural_log,
            convert_label_to_clicks.__name__: convert_label_to_clicks
            # Add more here
        }

    def add_fn(self, key, fn):
        """
        Add custom preprocessing function to the PreprocessingMap

        Parameters
        ----------
        key : str
            Name of the feature preprocessing function
        fn : function
            Function definition for the preprocessing function
        """
        self.key_to_fn[key] = fn

    def add_fns(self, keys_to_fns_dict):
        """
        Add custom preprocessing functions to the PreprocessingMap

        Parameters
        ----------
        keys_to_fns_dict : dict
            Dictionary of preprocessing functions to add to PreprocessingMap
        """
        self.key_to_fn.update(keys_to_fns_dict)

    def get_fns(self):
        """
        Get dictionary of functions in PreprocessingMap

        Returns
        -------
        dict
            Dictionary of preprocessing functions
        """
        return self.key_to_fn

    def get_fn(self, key):
        """
        Get preprocessing function from name

        Parameters
        ----------
        key : str
            Name of preprocessing function to fetch

        Returns
        -------
        function
            Function to preprocess feature corresponding to the key passed
        """
        return self.key_to_fn.get(key)

    def pop_fn(self, key):
        """
        Get preprocessing function from name and remove from PreprocessingMap

        Parameters
        ----------
        key : str
            Name of preprocessing function to fetch

        Returns
        -------
        function
            Function to preprocess feature corresponding to the key passed
        """
        self.key_to_fn.pop(key)


@tf.function
def preprocess_text(
    feature_tensor,
    remove_punctuation=False,
    to_lower=False,
    punctuation=string.punctuation,
    replace_with_whitespace=False,
):
    """
    String preprocessing function that removes punctuation and converts strings to lower case
    based on the arguments.

    Parameters
    feature_tensor : Tensor object
        input feature tensor of type tf.string
    remove_punctuation : bool
        Whether to remove punctuation characters from strings
    to_lower : bool
        Whether to convert string to lower case
    punctuation : str
        Punctuation characters to replace (a single string containing the character to remove
    replace_with_whitespace : bool
        if True punctuation will be replaced by whitespace (i.e. used as separator), note that
        leading and trailing whitespace will also be removed, as well as consecutive whitespaces.

    Returns
    -------
    Tensor object
        Processed string tensor

    Examples
    --------
    Input:
        >>> feature_tensor = "ABCabc123,,,"
        >>> remove_punctuation = True
        >>> to_lower = True
    Output:
        >>> "abcabc123"
    """
    if remove_punctuation:
        replacement = ""
        if replace_with_whitespace:
            replacement = " "
        punctuation_regex = (
            "[" + "".join([re.escape(c) for c in list(punctuation + replacement)]) + "]+"
        )
        feature_tensor = tf.strings.regex_replace(feature_tensor, punctuation_regex, replacement)

        if replace_with_whitespace:
            feature_tensor = tf.strings.strip(feature_tensor)

    if to_lower:
        feature_tensor = tf.strings.lower(feature_tensor)

    return feature_tensor


def get_one_hot_label_vectorizer(feature_info, file_io: FileIO):
    """
    Returns a tf function to convert categorical string labels to a one hot encoding.

    Parameters
    ----------
    feature_info : dict
        Dictionary representing the configuration parameters for the specific feature from the FeatureConfig.
        See categorical_indicator_with_vocabulary_file, here it is used to read a vocabulary file to
        create the one hot encoding.
    file_io: FileIO required to load the vocabulary file.

    Returns
    -------
    function
        Function that converts labels into one hot vectors

    Examples
    --------
    Input:
        >>> feature_tensor = ["abc", "xyz", "abc"]
        >>> vocabulary file
        >>>    abc -> 0
        >>>    xyz -> 1
        >>>    def -> 2
    Output:
        >>> [[1, 0, 0], [0, 1, 0], [1, 0, 0]]
    """
    label_str = tf.keras.Input(shape=(1,), dtype=tf.string)
    label_one_hot = categorical_indicator_with_vocabulary_file(label_str, feature_info, file_io)
    # FIXME: we should avoid use another keras Model here (we are wrapping two Keras models here, which cause issues at
    #  saving time).
    one_hot_vectorizer = tf.keras.Model(inputs=label_str, outputs=label_one_hot)

    @tf.function
    def one_hot_vectorize(feature_tensor):
        """
        Convert label tensor to one hot vector representation

        Parameters
        ----------
        feature_tensor : Tensor object
            input feature tensor of type tf.string

        Returns
        -------
            Tensor with numerical value corresponding to the one hot encoding from this string.
        """
        return tf.squeeze(one_hot_vectorizer(feature_tensor), axis=[0])

    return one_hot_vectorize


@tf.function
def split_and_pad_string(feature_tensor, split_char=",", max_length=20):
    """
    String preprocessing function that splits and pads a sequence based on the max_length.

    Parameters
    ----------
    feature_tensor : Tensor object
        Input feature tensor of type tf.string.
    split_char : str
        String separator to split the string input.
    max_length : int
        max length of the sequence produced after padding.

    Returns
    -------
    Tensor object
        processed float tensor

    Examples
    --------
    Input:
        >>> feature_tensor = "AAA,BBB,CCC"
        >>> split_char = ","
        >>> max_length = 5
    Output:
        >>> ['AAA', 'BBB', 'CCC', '', '']
    """
    tokens = tf.strings.split(feature_tensor, sep=split_char).to_tensor()
    padded_tokens = tf.image.pad_to_bounding_box(
        tf.expand_dims(tokens[:, :max_length], axis=-1),
        offset_height=0,
        offset_width=0,
        target_height=1,
        target_width=max_length,
    )
    padded_tokens = tf.squeeze(padded_tokens, axis=-1)
    return padded_tokens


@tf.function
def natural_log(feature_tensor, shift=1.0):
    """
    Compute the signed log of the feature_tensor

    Parameters
    ----------
    feature_tensor : Tensor object
        input feature tensor of type tf.float32
    shift : int
        floating point shift that is added to the feature tensor element wise before computing natural log
        (used to handle 0 values)

    Examples
    --------
    Input:
        >>> feature_tensor = [10, 0]
        >>> shift = 1
    Output:
        >>> [2.39, 0.]
    """
    return tf.math.log(tf.add(feature_tensor, tf.cast(tf.constant(shift), tf.float32)))

@tf.function
def convert_label_to_clicks(label_vector, dtype):
    """Convert the label vector to binary clicks. Documents with the maximum labels are considered clicked and receive
        label (1). Any other document is considered not clicked and receive label (0)
            Parameters
            ----------
            label_vector : tf tensor
                input label tensor of type label_dtype
            dtype : str
                Data type of the input label_vector


            Returns
            -------
            tf tensor
                converted clicks
    """

    typ = dtype
    if dtype == 'int':
        typ = 'int64'
    maximum = tf.reduce_max(label_vector)
    cond = tf.math.equal(label_vector, maximum)
    clicks = tf.dtypes.cast(cond, typ)
    return clicks

##########################################
# Add any new preprocessing functions here
##########################################



import unittest
from unittest.mock import patch
import json

import pandas as pd
from pandas import testing as pd_testing

from ml4ir.applications.ranking.model.metrics.metrics_helper import *


class ComputeSecondaryMetricsTest(unittest.TestCase):
    """Test suite for ml4ir.applications.ranking.model.metrics.metrics_helper"""

    def test_compute_secondary_label_metrics_1(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([10, 10, 10, 10, 10, 10, 10, 10, 1]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 1.,
                "test_label_failure_all": 0,
                "test_label_failure_any": 0,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 0,
                "test_label_failure_any_count": 0,
                "test_label_failure_any_fraction": 0.0}),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_2(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 1, 1, 1, 1, 1, 1, 1, 10]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.307,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 9,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 8,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_3(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([10, 10, 10, 1]),
            ranks=pd.Series(list(range(1, 5))),
            click_rank=4,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 1.,
                "test_label_failure_all": 0,
                "test_label_failure_any": 0,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 0,
                "test_label_failure_any_count": 0,
                "test_label_failure_any_fraction": 0.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_4(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 1, 1, 10]),
            ranks=pd.Series(list(range(1, 5))),
            click_rank=4,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.434,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 4,
                "test_label_failure_any_rank": 4,
                "test_label_failure_any_count": 3,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_5(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 1, 1, 5]),
            ranks=pd.Series(list(range(1, 5))),
            click_rank=4,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.514,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 4,
                "test_label_failure_any_rank": 4,
                "test_label_failure_any_count": 3,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_6(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([5, 5, 5, 10]),
            ranks=pd.Series(list(range(1, 5))),
            click_rank=4,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.474,
                "test_label_failure_all": 1.0,
                "test_label_failure_any": 1.0,
                "test_label_failure_all_rank": 4.0,
                "test_label_failure_any_rank": 4.0,
                "test_label_failure_any_count": 3.0,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_7(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 1, 1, 1, 5, 5, 5, 5, 10]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.329,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 9,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 8,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_8(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([5, 5, 5, 5, 1, 1, 1, 1, 10]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.361,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 9,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 8,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_9(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 5, 1, 5, 1, 5, 1, 5, 10]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.338,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 9,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 8,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_10(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([5, 1, 5, 1, 5, 1, 5, 1, 10]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.351,
                "test_label_failure_all": 1,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 9,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 8,
                "test_label_failure_any_fraction": 1.0
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_11(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 1, 1, 1, 10, 10, 10, 10, 5]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.546,
                "test_label_failure_all": 0,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 4,
                "test_label_failure_any_fraction": 0.5
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_12(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([10, 10, 10, 10, 1, 1, 1, 1, 5]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.999,
                "test_label_failure_all": 0,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 4,
                "test_label_failure_any_fraction": 0.5
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_13(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([1, 10, 1, 10, 1, 10, 1, 10, 5]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.678,
                "test_label_failure_all": 0,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 4,
                "test_label_failure_any_fraction": 0.5
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_14(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([10, 1, 10, 1, 10, 1, 10, 1, 5]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=9,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.867,
                "test_label_failure_all": 0,
                "test_label_failure_any": 1,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 9,
                "test_label_failure_any_count": 4,
                "test_label_failure_any_fraction": 0.5
            }),
            check_less_precise=True)

    def test_compute_secondary_label_metrics_invalid_click(self):
        computed_metrics = compute_secondary_label_metrics(
            secondary_label_values=pd.Series([10, 1, 10, 1, 10, 1, 10, 1, 5]),
            ranks=pd.Series(list(range(1, 10))),
            click_rank=15,
            secondary_label="test_label")
        pd_testing.assert_series_equal(
            pd.Series(computed_metrics),
            pd.Series({
                "test_label_NDCG": 0.867,
                "test_label_failure_all": 0,
                "test_label_failure_any": 0,
                "test_label_failure_all_rank": 0,
                "test_label_failure_any_rank": 0,
                "test_label_failure_any_count": 0,
                "test_label_failure_any_fraction": 0.
            }),
            check_less_precise=True)

    @patch("ml4ir.applications.ranking.model.metrics.metrics_helper.compute_secondary_label_metrics")
    def test_compute_secondary_labels_metrics_on_query_group(self, mock_compute_secondary_label_metrics):
        query_group = pd.DataFrame({
            "old_rank": [1, 2, 3, 4, 5],
            "new_rank": [3, 2, 1, 5, 4],
            "click": [0, 0, 1, 0, 0],
            "secondary_label_1": [5, 5, 5, 2, 2],
            "secondary_label_2": [3, 1, 2, 2, 5]
        })
        mock_compute_secondary_label_metrics.return_value = {}
        compute_secondary_labels_metrics_on_query_group(
            query_group=query_group,
            label_col="click",
            old_rank_col="old_rank",
            new_rank_col="new_rank",
            secondary_labels=["secondary_label_1", "secondary_label_2"])

        assert mock_compute_secondary_label_metrics.call_count == 2 * 2

        call_args = [args[1] for args in mock_compute_secondary_label_metrics.call_args_list]
        i = 0
        for secondary_label in ["secondary_label_1", "secondary_label_2"]:
            for state in ["old", "new"]:
                assert pd.Series.equals(
                    call_args[i]["secondary_label_values"], query_group[secondary_label])
                assert pd.Series.equals(call_args[i]["ranks"], query_group["{}_rank".format(state)])
                assert call_args[i]["click_rank"] == query_group[query_group["click"]
                                                                 == 1]["{}_rank".format(state)].values[0]
                assert call_args[i]["secondary_label"] == secondary_label
                assert call_args[i]["prefix"] == "{}_".format(state)

                i += 1

    def test_compute_secondary_labels_metrics_on_query_group_invalid_click(self):
        query_group = pd.DataFrame({
            "old_rank": [1, 2, 3, 4, 5],
            "new_rank": [3, 2, 1, 5, 4],
            "click": [0, 0, 0, 0, 0],
            "secondary_label_1": [5, 5, 5, 2, 2],
            "secondary_label_2": [3, 1, 2, 2, 5]
        })

        secondary_labels_metrics = compute_secondary_labels_metrics_on_query_group(
            query_group=query_group,
            label_col="click",
            old_rank_col="old_rank",
            new_rank_col="new_rank",
            secondary_labels=["secondary_label_1", "secondary_label_2"])
        assert secondary_labels_metrics.empty

    def test_compute_dcg(self):
        with self.subTest("Worst ordering of grade values"):
            self.assertTrue(np.isclose(compute_dcg([1., 2., 3.]), 4.261, atol=3))

        with self.subTest("Best ordering of grade values"):
            self.assertTrue(np.isclose(compute_dcg([3., 2., 1.]), 5.761, atol=3))

        with self.subTest("Equal grade values"):
            self.assertTrue(np.isclose(compute_dcg([1., 1., 1.]), 2.130, atol=3))

        with self.subTest("Zero grade values"):
            self.assertTrue(np.isclose(compute_dcg([0., 0., 0.]), 1.065, atol=3))

    def test_compute_ndcg(self):
        with self.subTest("Worst ordering of grade values"):
            self.assertTrue(np.isclose(compute_ndcg([1., 2., 3.]), 0.739, atol=3))

        with self.subTest("Best ordering of grade values"):
            self.assertTrue(np.isclose(compute_ndcg([3., 2., 1.]), 1., atol=3))

        with self.subTest("Equal grade values"):
            self.assertTrue(np.isclose(compute_ndcg([1., 1., 1.]), 1., atol=3))

        with self.subTest("Zero grade values"):
            self.assertTrue(np.isclose(compute_ndcg([0., 0., 0.]), 1., atol=3))



import pytest
import numpy as np
from ml4ir.applications.classification.tests.test_base import ClassificationTestBase


class ClassificationModelTest(ClassificationTestBase):
    """
    Test end-to-end model training for classification from model retrieved with
    :func:`~ml4ir.applications.classification.pipeline.ClassificationPipeline.get_relevance_model`
    """

    def test_csv_loss_metrics(self):
        """Test the loss from CSV data"""
        # Check if the loss, accuracy and top 5 accuracy on the test set is the same
        # Note that we don't check Precision which is not useful for this test model
        # Note that these numbers are different if you run it directly vs with docker-compose up
        expected_loss = 1.7
        tol = 0.2
        self.assertTrue(np.isclose(self.metrics_dict["loss"], expected_loss, atol=tol),
                        msg=f"Loss not in expected range."
                            f" Expected: {expected_loss} ± {tol}, Found: {self.metrics_dict['loss']}")

        expected_acc = 0.2
        tol = 0.05
        self.assertTrue(np.isclose(self.metrics_dict["categorical_accuracy"], expected_acc, atol=tol),
                        msg=f"Categorical_accuracy not in expected range."
                            f" Expected: {expected_acc} ± {tol}, Found: {self.metrics_dict['categorical_accuracy']}")

        expected_acc = 1.0
        tol = 0.01
        _metric = "top_5_categorical_accuracy"
        self.assertTrue(np.isclose(self.metrics_dict[_metric], expected_acc, atol=tol),
                        msg=f"Top5 Categorical_accuracy not in expected range."
                            f" Expected: {expected_acc} ± {tol}, Found: {self.metrics_dict[_metric]}")

        # Assert we predict for all the items
        expected_size_predictions = 200  # the same with data in test
        self.assertTrue(self.predictions.shape[0] == expected_size_predictions)

    def test_group_metrics_df(self):
        """
        Test the dimensions of the grouped metrics
        """
        # Metrics cardinality and names
        metrics = self.classification_pipeline.metrics_keys  # Metrics during training
        df = self.grouped_metrics
        self.assertTrue(df.metric.nunique() == len(metrics))  # number of metrics
        self.assertTrue(set(df.metric.unique()) == set(metrics))  # metrics per se
        # Grouped keys cardinality and names
        group_keys = self.classification_pipeline.feature_config.get_group_metrics_keys()
        group_names = [key['name'] for key in group_keys]
        unique_group_names = set(df.group_name.unique())
        self.assertTrue(unique_group_names == set(group_names))  # group_names
        # non trainable features that are group keys
        group_names_nt = [key['name'] for key in group_keys if not key['trainable']]
        for gk in group_names_nt:
            self.assertTrue(gk in unique_group_names)  # Assert they appear in the dataframe
            unique_metrics_gk = set(df.loc[df.group_name == gk].metric.unique())
            self.assertTrue(unique_metrics_gk == set(metrics))  # Assert each metric appears for them



import tensorflow as tf
from tensorflow.keras import losses
from tensorflow.keras import layers
from tensorflow.keras.losses import Reduction

from ml4ir.applications.ranking.model.losses.loss_base import ListwiseLossBase


class SoftmaxCrossEntropy(ListwiseLossBase):
    def get_loss_fn(self, **kwargs):
        """
        Define a softmax cross entropy loss

        Returns
        -------
        function
            Function to compute softmax cross entropy loss

        Notes
        -----
            Uses `mask` field to exclude padded records from contributing
            to the loss
        """
        cce = losses.CategoricalCrossentropy()
        mask = kwargs.get("mask")

        def _loss_fn(y_true, y_pred):
            """
            Shapes
            ------
            y_true : [batch_size, num_classes]
            y_pred : [batch_size, num_classes]
            mask : [batch_size, num_classes]
            """
            return cce(y_true, tf.math.multiply(y_pred, mask))

        return _loss_fn

    def get_final_activation_op(self, output_name):
        """
        Define a masked softmax activation function.
        This is one of the simplest and most effective loss functions
        for training ranking models with single click label.

        Ref -> https://dl.acm.org/doi/10.1145/3341981.3344221

        Parameters
        ----------
        output_name : str
            Name of the output to apply softmax activation on

        Returns
        -------
        function
            Function to compute masked softmax

        Notes
        -----
            Uses `mask` field to exclude padded records from contributing
            to the softmax activation
        """
        softmax_op = layers.Softmax(axis=-1, name=output_name)

        def masked_softmax(logits, mask):
            """
            NOTE:
            Tried to manually compute softmax with tf operations,
            but tf.keras.layers.Softmax() is more stable when working with
            cross_entropy layers
            """
            logits = tf.where(
                tf.equal(mask, tf.constant(1.0)), logits, tf.constant(tf.float32.min)
            )

            return softmax_op(logits)

        return masked_softmax


class RankOneListNet(SoftmaxCrossEntropy):
    def get_loss_fn(self, **kwargs):
        """
        Define a masked rank 1 ListNet loss.
        This loss is useful for multi-label classification when we have multiple
        click labels per document. This is because the loss breaks down the comparison
        between y_pred and y_true into individual binary assessments.

        Ref -> https://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/tr-2007-40.pdf

        Returns
        -------
        function
            Function to compute top 1 listnet loss

        Notes
        -----
            Uses `mask` field to exclude padded records from contributing
            to the loss
        """
        bce = losses.BinaryCrossentropy(reduction=Reduction.SUM)
        mask = kwargs.get("mask")

        def _loss_fn(y_true, y_pred):
            """
            Shapes
            ------
            y_true : [batch_size, num_classes]
            y_pred : [batch_size, num_classes]
            mask : [batch_size, num_classes]
            """
            batch_size = tf.cast(tf.shape(y_true)[0], tf.float32)

            # Mask the padded records
            y_true = tf.gather_nd(y_true, tf.where(tf.equal(mask, tf.constant(1.0))))
            y_pred = tf.gather_nd(y_pred, tf.where(tf.equal(mask, tf.constant(1.0))))

            # Reshape the tensors so that we sum the losses from each record
            y_true = tf.expand_dims(tf.squeeze(y_true), axis=-1)
            y_pred = tf.expand_dims(tf.squeeze(y_pred), axis=-1)

            # Scale the sum of losses down by number of queries in the batch
            return tf.math.divide(bce(y_true, y_pred), batch_size)

        return _loss_fn


import tensorflow as tf


def convert_score_to_rank(features, label, scores):
    """
    Convert scores for each record to rank within query

    Parameters
    ----------
    features: dict of tensors
        Dictionary of tensors that are the features used by the model
    label: Tensor
        Tensor object for the true label
    score: Tensor
        Tensor object containing the scores for each record within a query
        Shape -> [batch_size, sequence_size, 1]

    Returns
    -------
    Tensor
        Rank of each record within a query for all queries in the batch
        Shape -> [batch_size, sequence_size, 1]
    """
    return tf.add(
        tf.argsort(tf.argsort(scores, axis=-1, direction="DESCENDING", stable=True), stable=True),
        tf.constant(1),
    )



from ml4ir.base.features.feature_fns import categorical as categorical_fns
from ml4ir.base.features.feature_fns import sequence as sequence_fns
from ml4ir.base.features.feature_fns import tf_native as tf_native_fns
from ml4ir.base.config.keys import SequenceExampleTypeKey
from ml4ir.base.tests.test_base import RelevanceTestBase

import tensorflow as tf
import numpy as np


class FeatureLayerTest(RelevanceTestBase):
    def test_bytes_sequence_to_encoding_bilstm(self):
        """
        Asserts the conversion of a string tensor to its corresponding sequence encoding
        obtained through the bytes_sequence_to_encoding_bilstm function
        Works by converting each string into a bytes sequence and then
        passing it through a biLSTM.

        The embedding and encoding dimensions are controlled by the feature_info
        """
        max_length = 20
        embedding_size = 128
        encoding_size = 512
        feature_info = {
            "feature_layer_info": {
                "type": "numeric",
                "fn": "bytes_sequence_to_encoding_bilstm",
                "args": {
                    "encoding_type": "bilstm",
                    "encoding_size": encoding_size,
                    "embedding_size": embedding_size,
                    "max_length": max_length,
                },
            },
            "tfrecord_type": SequenceExampleTypeKey.CONTEXT,
        }

        # Define an input string tensor
        string_tensor = [["abc"], ["xyz"], ["123"]]
        sequence_encoding = sequence_fns.bytes_sequence_to_encoding_bilstm(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting encoding based on the feature_info
        assert sequence_encoding.shape[0] == len(string_tensor)
        assert sequence_encoding.shape[1] == 1
        assert sequence_encoding.shape[2] == encoding_size

    def test_categorical_embedding_to_encoding_bilstm(self):
        """
        Asserts the conversion of a string tensor to its corresponding sequence encoding
        obtained through the categorical_embedding_to_encoding_bilstm function
        Works by converting each string into a bytes sequence and then
        passing it through a biLSTM.

        The embedding and encoding dimensions are controlled by the feature_info
        """
        embedding_size = 128
        encoding_size = 512
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "type": "numeric",
                "fn": "categorical_embedding_to_encoding_bilstm",
                "args": {
                    "vocabulary_file": "ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv",
                    "encoding_type": "bilstm",
                    "encoding_size": encoding_size,
                    "embedding_size": embedding_size,
                },
            },
            "tfrecord_type": SequenceExampleTypeKey.CONTEXT,
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            [[["AAA"]], [["BBB"]], [["AAA"]], [["CCC"]], [["out_of_vocabulary"]]]
        )
        sequence_encoding = categorical_fns.categorical_embedding_to_encoding_bilstm(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting encoding based on the feature_info
        assert sequence_encoding.shape[0] == len(string_tensor)
        assert sequence_encoding.shape[1] == 1
        assert sequence_encoding.shape[2] == encoding_size

        # Strings 0 and 2 should result in the same embedding because they are the same
        assert tf.reduce_all(tf.equal(sequence_encoding[0], sequence_encoding[2]))
        assert not tf.reduce_all(tf.equal(sequence_encoding[0], sequence_encoding[1]))
        assert not tf.reduce_all(tf.equal(sequence_encoding[3], sequence_encoding[4]))
        assert not tf.reduce_all(tf.equal(sequence_encoding[1], sequence_encoding[4]))

    def test_categorical_embedding_to_encoding_bilstm_file_truncation(self):
        """
        Asserts the conversion of a string tensor to its corresponding sequence encoding
        obtained through the categorical_embedding_to_encoding_bilstm function
        Works by converting each string into a bytes sequence and then
        passing it through a biLSTM.

        The embedding and encoding dimensions are controlled by the feature_info
        """
        max_length = 2
        embedding_size = 128
        encoding_size = 512
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "type": "numeric",
                "fn": "categorical_embedding_to_encoding_bilstm",
                "args": {
                    "vocabulary_file": "ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv",
                    "encoding_type": "bilstm",
                    "encoding_size": encoding_size,
                    "embedding_size": embedding_size,
                    "max_length": max_length,
                },
            },
            "tfrecord_type": SequenceExampleTypeKey.CONTEXT,
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            [[["AAA"]], [["BBB"]], [["AAA"]], [["CCC"]], [["out_of_vocabulary"]]]
        )
        sequence_encoding = categorical_fns.categorical_embedding_to_encoding_bilstm(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting encoding based on the feature_info
        assert sequence_encoding.shape[0] == len(string_tensor)
        assert sequence_encoding.shape[1] == 1
        assert sequence_encoding.shape[2] == encoding_size

        assert tf.reduce_all(tf.equal(sequence_encoding[0], sequence_encoding[2]))
        assert not tf.reduce_all(tf.equal(sequence_encoding[0], sequence_encoding[1]))
        assert tf.reduce_all(tf.equal(sequence_encoding[3], sequence_encoding[4]))
        assert not tf.reduce_all(tf.equal(sequence_encoding[1], sequence_encoding[4]))

    def test_categorical_embedding_to_encoding_bilstm_sequence_of_words(self):
        """
        Check that categorical_embedding_to_encoding_bilstm function does not fail with sequences of word (in a single
        row).
        """
        embedding_size = 32
        encoding_size = 64
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "type": "numeric",
                "fn": "categorical_embedding_to_encoding_bilstm",
                "args": {
                    "vocabulary_file": "ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv",
                    "encoding_type": "bilstm",
                    "encoding_size": encoding_size,
                    "embedding_size": embedding_size,
                    "dropout_rate": 0.2,
                },
            },
            "tfrecord_type": SequenceExampleTypeKey.CONTEXT,
        }

        string_tensor = tf.constant([[["AAA", "BBB", "out_of_vocabulary"]]])

        categorical_fns.categorical_embedding_to_encoding_bilstm(
            string_tensor, feature_info, self.file_io
        )

    def test_categorical_embedding_to_encoding_bilstm_oov_mapping_with_dropout(self):
        """
        Asserts "out of vocabulary" words are mapped to the same value when dropout_rate is used,
        while word "in vocabulary" are mapped to their own embeddings.
        """
        embedding_size = 128
        encoding_size = 512
        vocabulary_file = (
            "ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv"
        )

        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "type": "numeric",
                "fn": "categorical_embedding_to_encoding_bilstm",
                "args": {
                    "vocabulary_file": vocabulary_file,
                    "encoding_type": "bilstm",
                    "encoding_size": encoding_size,
                    "embedding_size": embedding_size,
                    "dropout_rate": 0.2,
                },
            },
            "tfrecord_type": SequenceExampleTypeKey.CONTEXT,
        }

        df_vocabulary = self.file_io.read_df(vocabulary_file)
        vocabulary = df_vocabulary.values.flatten().tolist()

        # Define a batch where each row is a single word from the vocabulary (so that we can later compare each word
        # encoding)
        words = vocabulary + ["out_of_vocabulary", "out_of_vocabulary2"]
        string_tensor = tf.constant([[[word]] for word in words])

        # If the embedding lookup input dimension were computed incorrectly, the following call would fail with:
        # tensorflow.python.framework.errors_impl.InvalidArgumentError: indices[0,0,7] = 8 is not in [0, 8)
        sequence_encoding = categorical_fns.categorical_embedding_to_encoding_bilstm(
            string_tensor, feature_info, self.file_io
        )

        # Strings 1 and 2 should result in the same embedding because they are both OOV
        # Check that each word, except the OOV ones are mapped to different encodings (actually different embeddings,
        # but the encoding on top should be equivalent as each row is a single word).
        for word1_position in range(0, len(vocabulary)):
            # +1 to include one OOV word
            for word2_position in range(word1_position + 1, len(vocabulary) + 1):
                if tf.reduce_all(
                    tf.equal(sequence_encoding[word1_position], sequence_encoding[word2_position])
                ):
                    word1, word2 = words[word1_position], words[word2_position]
                    self.assertFalse(
                        tf.reduce_all(
                            tf.equal(
                                sequence_encoding[word1_position],
                                sequence_encoding[word2_position],
                            )
                        ),
                        msg="Check that non-OOV words map to different encodings, and not also not to OOV:"
                        "{} vs. {}".format(word1, word2),
                    )

        self.assertTrue(
            tf.reduce_all(tf.equal(sequence_encoding[-1], sequence_encoding[-2])),
            msg="Check that OOV words map to the same encodings",
        )
        assert not tf.reduce_all(tf.equal(sequence_encoding[0], sequence_encoding[1]))

    def test_categorical_embedding_with_hash_buckets(self):
        """
        Goal:
        Convert a categorical string tensor into a categorical embedding
        Works by converting the string into n integers by hashing and then
        converting the integers into n embeddings.

        The embedding dimensions, buckets, etc are controlled by the feature_info
        """
        num_hash_buckets = 4
        hash_bucket_size = 64
        embedding_size = 32
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_embedding_with_hash_buckets",
                "args": {
                    "num_hash_buckets": num_hash_buckets,
                    "hash_bucket_size": hash_bucket_size,
                    "embedding_size": embedding_size,
                    "merge_mode": "concat",
                },
            },
        }

        # Define an input string tensor
        string_tensor = ["domain_0", "domain_1", "domain_0"]

        categorical_embedding = categorical_fns.categorical_embedding_with_hash_buckets(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting embedding
        assert categorical_embedding.shape[0] == len(string_tensor)
        assert categorical_embedding.shape[1] == 1
        assert categorical_embedding.shape[2] == num_hash_buckets * embedding_size

        # Strings 0 and 2 should result in the same embedding because they are the same
        assert tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[2]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[1]))

    def test_categorical_embedding_with_indices(self):
        """
        Asserts the conversion of integer categorical indices tensor into categorical embeddings

        The embeddding dimensions are controlled by the feature_info
        """
        num_buckets = 8
        embedding_size = 64
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_embedding_with_indices",
                "args": {
                    "num_buckets": num_buckets,
                    "embedding_size": embedding_size,
                    "default_value": 0,
                },
            },
        }

        # Define an input int tensor
        index_tensor = [0, 1, 2, 1, 10]

        categorical_embedding = categorical_fns.categorical_embedding_with_indices(
            index_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting embedding
        assert categorical_embedding.shape[0] == len(index_tensor)
        assert categorical_embedding.shape[1] == 1
        assert categorical_embedding.shape[2] == embedding_size

        # Assert equality of embeddings with same indices
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[1]))
        assert tf.reduce_all(tf.equal(categorical_embedding[1], categorical_embedding[3]))
        assert tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[-1]))

    def test_categorical_embedding_with_vocabulary_file_with_ids(self):
        """
        Asserts the conversion of a categorical string tensor into a categorical embedding
        Works by converting the string into indices using a vocabulary file and then
        converting the indices into categorical embeddings

        The embedding dimensions, buckets, etc are controlled by the feature_info
        """
        embedding_size = 32
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_embedding_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/ranking/tests/data/configs/domain_name_vocab.csv",
                    "embedding_size": embedding_size,
                    "default_value": -1,
                    "num_oov_buckets": 1,
                },
            },
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            ["domain_0", "domain_1", "domain_0", "domain_2", "domain_10", "domain_11"]
        )

        categorical_embedding = categorical_fns.categorical_embedding_with_vocabulary_file(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting embedding
        assert categorical_embedding.shape[0] == len(string_tensor)
        assert categorical_embedding.shape[1] == 1
        assert categorical_embedding.shape[2] == embedding_size

        # Strings 0 and 2 should result in the same embedding because they are the same
        assert tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[2]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[1]))
        assert tf.reduce_all(tf.equal(categorical_embedding[4], categorical_embedding[5]))

        # Strings domain_0 and domain_2 should result in the same embedding because they are mapped to the same ID
        assert tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[3]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[3], categorical_embedding[4]))

    def test_categorical_embedding_with_vocabulary_file_without_ids(self):
        """
        Asserts the conversion of a categorical string tensor into a categorical embedding
        Works by converting the string into indices using a vocabulary file and then
        converting the indices into categorical embeddings

        The embedding dimensions, buckets, etc are controlled by the feature_info
        """
        embedding_size = 32
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_embedding_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/ranking/tests/data/configs/domain_name_vocab_no_id.csv",
                    "embedding_size": embedding_size,
                    "default_value": -1,
                    "num_oov_buckets": 1,
                },
            },
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            ["domain_0", "domain_1", "domain_0", "domain_2", "domain_10", "domain_11"]
        )

        categorical_embedding = categorical_fns.categorical_embedding_with_vocabulary_file(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting embedding
        assert categorical_embedding.shape[0] == len(string_tensor)
        assert categorical_embedding.shape[1] == 1
        assert categorical_embedding.shape[2] == embedding_size

        # Strings 0 and 2 should result in the same embedding because they are the same
        assert tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[2]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[1]))
        assert tf.reduce_all(tf.equal(categorical_embedding[4], categorical_embedding[5]))

        # Strings domain_0 and domain_2 should NOT result in the same embedding because they use a default one-to-one vocabulary mapping
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[3]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[3], categorical_embedding[4]))

    def test_categorical_embedding_with_vocabulary_file_with_ids_and_dropout(self):
        """
        Asserts the conversion of a categorical string tensor into an embedding representation
        Works by converting the string into indices using a vocabulary file and then dropping these indices into the OOV index at dropout_rate rate.

        The embedding size, dropout_rate are controlled by feature_info
        """
        embedding_size = 4
        dropout_rate = 0.999
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_embedding_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/ranking/tests/data/configs/domain_name_vocab.csv",
                    "embedding_size": embedding_size,
                    "dropout_rate": dropout_rate,
                },
            },
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            ["domain_0", "domain_1", "domain_0", "domain_2", "domain_10", "domain_11"]
        )

        value_error_raised = False
        try:
            categorical_fns.categorical_embedding_with_vocabulary_file_and_dropout(
                string_tensor, feature_info, self.file_io
            )
        except ValueError:
            # Should throw error as method does not work with IDs containing 0
            value_error_raised = True

        assert value_error_raised

    def test_categorical_embedding_with_vocabulary_file_without_ids_and_dropout(self):
        """
        Asserts the conversion of a categorical string tensor into an embedding representation
        Works by converting the string into indices using a vocabulary file and then dropping these indices into the OOV index at dropout_rate rate.

        The embedding size, dropout_rate are controlled by feature_info
        """
        embedding_size = 4
        dropout_rate = 0.999
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_embedding_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/ranking/tests/data/configs/domain_name_vocab_no_id.csv",
                    "embedding_size": embedding_size,
                    "dropout_rate": dropout_rate,
                },
            },
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            ["domain_0", "domain_1", "domain_0", "domain_2", "domain_10", "domain_11"]
        )

        categorcial_tensor = tf.keras.Input(shape=(1,), dtype=tf.string)
        embedding_tensor = categorical_fns.categorical_embedding_with_vocabulary_file_and_dropout(
            categorcial_tensor, feature_info, self.file_io
        )
        model = tf.keras.Model(categorcial_tensor, embedding_tensor)

        categorical_embedding = model(string_tensor, training=False)

        # Assert the right shapes of the resulting embedding
        assert categorical_embedding.shape[0] == len(string_tensor)
        assert categorical_embedding.shape[1] == 1
        assert categorical_embedding.shape[2] == embedding_size

        # Strings 0 and 2 should result in the same embedding because they are the same
        assert tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[2]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[1]))
        assert tf.reduce_all(tf.equal(categorical_embedding[4], categorical_embedding[5]))

        # Strings domain_0 and domain_2 should NOT result in the same embedding because they use a default one-to-one vocabulary mapping
        assert not tf.reduce_all(tf.equal(categorical_embedding[0], categorical_embedding[3]))
        assert not tf.reduce_all(tf.equal(categorical_embedding[3], categorical_embedding[4]))

        categorical_embedding = model(string_tensor, training=True)
        # Since dropout_rate is set to 0.999, all categorical indices
        # should be masked to OOV index and thus the embeddings should be the same
        assert tf.reduce_all(tf.equal(categorical_embedding, categorical_embedding[0]))

    def test_categorical_indicator_with_vocabulary_file_with_ids(self):
        """
        Asserts the conversion of a categorical string tensor into a one-hot representation
        Works by converting the string into indices using a vocabulary file and then
        converting the indices into one-hot vectors

        The one-hot vector dimensions, buckets, etc are controlled by the feature_info
        """
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_indicator_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/ranking/tests/data/configs/domain_name_vocab.csv",
                    "num_oov_buckets": 1,
                },
            },
            "default_value": "",
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            ["domain_0", "domain_1", "domain_0", "domain_2", "domain_10", "domain_11"]
        )

        categorical_one_hot = categorical_fns.categorical_indicator_with_vocabulary_file(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting one-hot vector
        assert categorical_one_hot.shape[0] == len(string_tensor)
        assert categorical_one_hot.shape[1] == 1
        assert categorical_one_hot.shape[2] == 6
        assert tf.reduce_all(tf.squeeze(tf.reduce_sum(categorical_one_hot, axis=2)) == 1.0)

        # Strings 0 and 2 should result in the same one-hot vector because they are the same
        assert tf.reduce_all(tf.equal(categorical_one_hot[0], categorical_one_hot[2]))
        assert not tf.reduce_all(tf.equal(categorical_one_hot[0], categorical_one_hot[1]))
        assert tf.reduce_all(tf.equal(categorical_one_hot[4], categorical_one_hot[5]))

        # Strings domain_0 and domain_2 should result in the same one-hot vector because they are mapped to the same ID
        assert tf.reduce_all(tf.equal(categorical_one_hot[0], categorical_one_hot[3]))
        assert not tf.reduce_all(tf.equal(categorical_one_hot[3], categorical_one_hot[4]))

    def test_categorical_indicator_with_vocabulary_file_without_ids(self):
        """
        Asserts the conversion of a categorical string tensor into a one-hot representation
        Works by converting the string into indices using a vocabulary file and then
        converting the indices into one-hot vectors

        The one-hot vector dimensions, buckets, etc are controlled by the feature_info
        """
        feature_info = {
            "name": "categorical_variable",
            "feature_layer_info": {
                "fn": "categorical_indicator_with_vocabulary_file",
                "args": {
                    "vocabulary_file": "ml4ir/applications/ranking/tests/data/configs/domain_name_vocab_no_id.csv",
                    "num_oov_buckets": 1,
                },
            },
            "default_value": "",
        }

        # Define an input string tensor
        string_tensor = tf.constant(
            ["domain_0", "domain_1", "domain_0", "domain_2", "domain_10", "domain_11"]
        )

        categorical_one_hot = categorical_fns.categorical_indicator_with_vocabulary_file(
            string_tensor, feature_info, self.file_io
        )

        # Assert the right shapes of the resulting one-hot vector
        assert categorical_one_hot.shape[0] == len(string_tensor)
        assert categorical_one_hot.shape[1] == 1
        assert categorical_one_hot.shape[2] == 6
        assert tf.reduce_all(tf.squeeze(tf.reduce_sum(categorical_one_hot, axis=2)) == 1.0)

        # Strings 0 and 2 should result in the same one-hot vector because they are the same
        assert tf.reduce_all(tf.equal(categorical_one_hot[0], categorical_one_hot[2]))
        assert not tf.reduce_all(tf.equal(categorical_one_hot[0], categorical_one_hot[1]))
        assert tf.reduce_all(tf.equal(categorical_one_hot[4], categorical_one_hot[5]))

        # Strings domain_0 and domain_2 should NOT result in the same one-hot vector because they use a default one-to-one vocabulary mapping
        assert not tf.reduce_all(tf.equal(categorical_one_hot[0], categorical_one_hot[3]))
        assert not tf.reduce_all(tf.equal(categorical_one_hot[3], categorical_one_hot[4]))

    def test_global_1d_pooling(self):
        """
        Unit test the global 1D pooling feature function on sequence features

        Checks the right output shapes produced and the values generated
        """
        feature_tensor = tf.reshape(tf.constant(range(30), dtype=tf.float32), (2, 5, 3))
        pooled_tensor = sequence_fns.global_1d_pooling(
            feature_tensor=feature_tensor,
            feature_info={
                "name": "f",
                "feature_layer_info": {
                    "args": {"fns": ["sum", "mean", "max", "min", "count_nonzero"]}
                },
            },
            file_io=None,
        )

        assert pooled_tensor.shape == (2, 5, 5)
        assert (
            pooled_tensor.numpy()
            == [
                [
                    [3.0, 1.0, 2.0, 0.0, 2.0],
                    [12.0, 4.0, 5.0, 3.0, 3.0],
                    [21.0, 7.0, 8.0, 6.0, 3.0],
                    [30.0, 10.0, 11.0, 9.0, 3.0],
                    [39.0, 13.0, 14.0, 12.0, 3.0],
                ],
                [
                    [48.0, 16.0, 17.0, 15.0, 3.0],
                    [57.0, 19.0, 20.0, 18.0, 3.0],
                    [66.0, 22.0, 23.0, 21.0, 3.0],
                    [75.0, 25.0, 26.0, 24.0, 3.0],
                    [84.0, 28.0, 29.0, 27.0, 3.0],
                ],
            ]
        ).all()

        # Test function with padded values
        # Here, we mask all the even values
        padded_val = -1.0
        feature_tensor_with_mask = tf.where(feature_tensor % 2 == 0, feature_tensor, padded_val)
        pooled_tensor_with_mask = sequence_fns.global_1d_pooling(
            feature_tensor=feature_tensor_with_mask,
            feature_info={
                "name": "f",
                "feature_layer_info": {
                    "args": {
                        "fns": ["sum", "mean", "max", "min", "count_nonzero"],
                        "padded_val": padded_val,
                        "masked_max_val": 100.0,
                    }
                },
            },
            file_io=None,
        )
        assert pooled_tensor_with_mask.shape == (2, 5, 5)
        assert not (pooled_tensor_with_mask.numpy() == pooled_tensor.numpy()).all()
        assert (
            pooled_tensor_with_mask.numpy()
            == [
                [
                    [2.0, 1.0, 2.0, 0.0, 1.0],
                    [4.0, 4.0, 4.0, 4.0, 1.0],
                    [14.0, 7.0, 8.0, 6.0, 2.0],
                    [10.0, 10.0, 10.0, 10.0, 1.0],
                    [26.0, 13.0, 14.0, 12.0, 2.0],
                ],
                [
                    [16.0, 16.0, 16.0, 16.0, 1.0],
                    [38.0, 19.0, 20.0, 18.0, 2.0],
                    [22.0, 22.0, 22.0, 22.0, 1.0],
                    [50.0, 25.0, 26.0, 24.0, 2.0],
                    [28.0, 28.0, 28.0, 28.0, 1.0],
                ],
            ]
        ).all()

        # Test empty pooling fn list
        found_value_error = False
        try:
            pooled_tensor = sequence_fns.global_1d_pooling(
                feature_tensor=feature_tensor,
                feature_info={
                    "name": "f",
                    "feature_layer_info": {"args": {"fns": [], "padded_val": padded_val}},
                },
                file_io=None,
            )
        except ValueError:
            found_value_error = True
        assert found_value_error

        # Test invalid pooling fn
        found_key_error = False
        try:
            pooled_tensor = sequence_fns.global_1d_pooling(
                feature_tensor=feature_tensor,
                feature_info={
                    "name": "f",
                    "feature_layer_info": {"args": {"fns": ["invalid"], "padded_val": padded_val}},
                },
                file_io=None,
            )
        except KeyError:
            found_key_error = True
        assert found_key_error

    def test_tf_native_op(self):
        """
        Unit test the tf_native_op feature function

        Checks the right output shapes produced and the values generated
        """
        input_tensor = np.abs(np.random.randn(32, 4, 2))

        actual_tensor = tf_native_fns.tf_native_op(
                feature_tensor=input_tensor,
                feature_info={
                    "name": "f",
                    "feature_layer_info": {
                        "args": {
                            "ops": [
                                {"fn": "tf.math.add", "args": {"y": 1.}},
                                {"fn": "tf.math.log"}
                            ]
                        }
                    }
                },
                file_io=None
            )
        expected_tensor = tf.expand_dims(tf.math.log(tf.math.add(input_tensor, 1.)), axis=-1)

        assert tf.reduce_all(actual_tensor == expected_tensor)



from tensorflow.keras.metrics import Metric

from ml4ir.applications.classification.config.keys import MetricKey
from ml4ir.applications.classification.model.metrics.metrics_impl import (
    CategoricalAccuracy,
    Top5CategoricalAccuracy,
)


def get_metric(metric_key: str) -> Metric:
    """
    Factory method to get Metric class

    Parameters
    ----------
    metric_key : str
        Name of the metric class to retrieve

    Returns
    -------
    Metric class
        Class defining the metric computation logic
    """
    if metric_key == MetricKey.CATEGORICAL_ACCURACY:
        return CategoricalAccuracy
    elif metric_key == MetricKey.TOP_5_CATEGORICAL_ACCURACY:
        return Top5CategoricalAccuracy
    else:
        return metric_key



import os
import sys
from typing import Optional

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow import data
from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.model.relevance_model import RelevanceModelConstants


class ClassificationModel(RelevanceModel):
    """Inherits from a Relevance model
    to create a classification model to
    create classification evaluate and predict
    methods."""

    def evaluate(
        self,
        test_dataset: data.TFRecordDataset,
        inference_signature: str = None,
        additional_features: dict = {},
        group_metrics_min_queries: int = 50,
        logs_dir: Optional[str] = None,
        logging_frequency: int = 25,
        compute_intermediate_stats: bool = True,
    ):
        """
        Evaluate the Classification Model
        To avoid evaluating once on the the whole test
        dataset, we calculate the metrics incrementally using
        the batch size the user defined. Currently, we have a hacky
        way to get the batch size since it is not part of the
        `relevance_model` or `classification_model`.

        Parameters
        ----------
        test_dataset: an instance of tf.data.dataset
        inference_signature : str, optional
            If using a SavedModel for prediction, specify the inference signature to be used for computing scores
        additional_features : dict, optional
            Dictionary containing new feature name and function definition to
            compute them. Use this to compute additional features from the scores.
            For example, converting ranking scores for each document into ranks for
            the query
        group_metrics_min_queries : int, optional
            Minimum count threshold per group to be considered for computing
            groupwise metrics
        logs_dir : str, optional
            Path to directory to save logs
        logging_frequency : int
            Value representing how often(in batches) to log status
        compute_intermediate_stats : bool
            Determines if group metrics and other intermediate stats on the test set should be computed

        Returns
        -------
        df_overall_metrics : `pd.DataFrame` object
            `pd.DataFrame` containing overall metrics
        df_groupwise_metrics : `pd.DataFrame` object
            `pd.DataFrame` containing groupwise metrics if
            group_metric_keys are defined in the FeatureConfig
        metrics_dict : dict
            metrics as a dictionary of metric names mapping to values

        Notes
        -----
        You can directly do a `model.evaluate()` only if the keras model is compiled
        """

        if not self.is_compiled:
            return NotImplementedError
        group_metrics_keys = self.feature_config.get_group_metrics_keys()
        metrics_dict = self.model.evaluate(test_dataset)
        metrics_dict = dict(zip(self.model.metrics_names, metrics_dict))
        if compute_intermediate_stats:
            self.logger.info("Computing grouped metrics.")
            self.logger.warning("Warning: currently, group-wise metric computation "
                                "collects the test data and predictions "
                                "in memory in order to perform the groupBy operations. "
                                "With large test datasets, it can lead in OOM issues.")
            predictions = self.predict(test_dataset,
                                       inference_signature=inference_signature,
                                       additional_features=additional_features,
                                       logs_dir=logs_dir,
                                       logging_frequency=logging_frequency)
            global_metrics = []  # group_name, metric, value
            grouped_metrics = []
            # instead of calculating measure with a single update_state (can result in a call with
            # thousands of examples at once, we use the same batch size used during training.
            # Helps prevent OOM issues.

            '''
            type(test_dataset._input_dataset)
            Out[4]: tensorflow.python.data.ops.dataset_ops.ShardDataset
            type(test_dataset)
            Out[10]: tensorflow.python.data.ops.dataset_ops.BatchDataset
            
            type(test_dataset)
            Out[3]: tensorflow.python.data.ops.dataset_ops.PrefetchDataset
            type(test_dataset._input_dataset)
            Out[4]: tensorflow.python.data.ops.dataset_ops.BatchDataset
            '''
            batch_size = test_dataset._input_dataset._batch_size.numpy()  # Hacky way to get batch_size
            # Letting metrics in the outer loop to avoid tracing
            for metric in self.model.metrics:
                global_metrics.append(
                    self.calculate_metric_on_batch(metric, predictions, batch_size))
                self.logger.info(f"Global metric {metric.name} completed."
                                 f" Score: {global_metrics[-1]['value']}")

                for group_ in group_metrics_keys:  # Calculate metrics for group metrics
                    for name, group in predictions.groupby(group_['name']):
                        self.logger.info(f"Per feature metric {metric.name}."
                                         f" Feature: {group_['name']}, value: {name}")
                        if group.shape[0] >= group_metrics_min_queries:
                            grouped_metrics.append(self.calculate_metric_on_batch(metric,
                                                                                  group,
                                                                                  batch_size,
                                                                                  group_['name'],
                                                                                  name))
            global_metrics = pd.DataFrame(global_metrics)
            grouped_metrics = pd.DataFrame(grouped_metrics).sort_values(by='size')
            if logs_dir:
                self.file_io.write_df(
                    grouped_metrics,
                    outfile=os.path.join(logs_dir, RelevanceModelConstants.GROUP_METRICS_CSV_FILE),
                    index=False
                )
                self.file_io.write_df(
                    global_metrics,
                    outfile=os.path.join(logs_dir, RelevanceModelConstants.METRICS_CSV_FILE),
                    index=False
                )
                self.logger.info(f"Evaluation Results written at: {logs_dir}")
            return global_metrics, grouped_metrics, metrics_dict
        if logs_dir:
            global_metrics = pd.DataFrame.from_dict(metrics_dict, orient='index', columns=["value"])
            global_metrics = global_metrics.reset_index().rename(columns={'index': 'metric'})
            self.file_io.write_df(
                global_metrics,
                outfile=os.path.join(logs_dir, RelevanceModelConstants.METRICS_CSV_FILE),
                index=False
            )
            self.logger.info(f"Evaluation Results written at: {logs_dir}")
        return None, None, metrics_dict

    @staticmethod
    def get_chunks_from_df(dataframe, size):
        """
        Given a pd.DataFrame, creates batches of size `size`.
        Instead of having a metric applied on big data once, this
        ensures we are using batches of examples during test too.
        """
        for pos in range(0, len(dataframe), size):
            yield dataframe.iloc[pos:pos + size]

    def calculate_metric_on_batch(self, metric, predictions, batch_size, group_name=None, group_key=None):
        """
        Given a metric and a dataframe with `predictions`, it iterates the dataframe in
        using `batch_size` and updates the metric. Once the dataframe is fully iterated,
        the score of the metric is returned as a dictionary containing
        {"group_name": group_name,
         "group_key": group_key,
         "metric": metric.name,
         "value": metric.result().numpy(),
         "size": predictions.shape[0]
        }
        When group_name and group_key are not provided, they are exluded from the dict.

        Parameters
        ----------
        metric : an instance of tf.keras.metrics
        predictions : pd.DataFrame
        batch_size: int
        group_name : str, optional
            The name to be used in the returned dictionary
        group_key : str, optional
            A key to be used in the returned dictionary

        Returns
        -------
        A dictionary with the metric scores for the given data.
        """
        label_name = self.feature_config.get_label()['name']
        output_name = self.output_name
        metric.reset_states()
        for chunk in self.get_chunks_from_df(predictions, batch_size):
            metric.update_state(tf.constant(chunk[label_name].values.tolist(), dtype=tf.float32),
                                tf.constant(chunk[output_name].values.tolist(), dtype=tf.float32))
        if group_name:
            return {"group_name": group_name,
                    "group_key": group_key,
                    "metric": metric.name,
                    "value": metric.result().numpy(),
                    "size": predictions.shape[0]}
        return {"metric": metric.name,
                "value": metric.result().numpy(),
                "size": predictions.shape[0]}

    def predict(
            self,
            test_dataset: data.TFRecordDataset,
            inference_signature: str = "serving_default",
            additional_features: dict = {},
            logs_dir: Optional[str] = None,
            logging_frequency: int = 25,
    ):
        """
        Predict the scores on the test dataset using the trained model

        Parameters
        ----------
        test_dataset : `Dataset` object
            `Dataset` object for which predictions are to be made
        inference_signature : str, optional
            If using a SavedModel for prediction, specify the inference signature to be used for computing scores
        additional_features : dict, optional
            Dictionary containing new feature name and function definition to
            compute them. Use this to compute additional features from the scores.
            For example, converting ranking scores for each document into ranks for
            the query
        logs_dir : str, optional
            Path to directory to save logs
        logging_frequency : int
            Value representing how often(in batches) to log status

        Returns
        -------
        `pd.DataFrame`
            pandas DataFrame containing the predictions on the test dataset
            made with the `RelevanceModel`
        """
        if logs_dir:
            outfile = os.path.join(logs_dir, RelevanceModelConstants.MODEL_PREDICTIONS_CSV_FILE)
            # Delete file if it exists
            self.file_io.rm_file(outfile)
        predictions_df = self._create_prediction_dataframe(logging_frequency,
                                                           test_dataset)
        predictions_ = np.squeeze(self.model.predict(test_dataset))
        # Below, avoid doing predictions.tolist() as it explodes the memory
        # tolist() will create a list of lists, which consumes more memory
        # than a list on numpy arrays
        predictions_df[self.output_name] = [x for x in predictions_]
        if logs_dir:
            np.set_printoptions(formatter={'all':lambda x: str(x.decode('utf-8')) if isinstance(x, bytes) else str(x)},
                                linewidth=sys.maxsize, threshold=sys.maxsize)  # write the full vector in the csv not ...
            for col in predictions_df.columns:
                if isinstance(predictions_df[col].values[0], bytes):
                    predictions_df[col] = predictions_df[col].str.decode('utf8')
            predictions_df.to_csv(outfile, mode="w", header=True, index=False)
            self.logger.info(f"Model predictions written to: {outfile}")
        return predictions_df

    def _create_prediction_dataframe(self, logging_frequency, test_dataset):
        """
        Iterates through the test data and collects for each
        data point the information to be logged.
        """
        predictions = {}  # keys are features we logs, values lists
        # containing the values of this feature for each batch
        label_name = self.feature_config.get_label()["name"]
        features_to_log = [f.get("node_name", f["name"]) for f in
                           self.feature_config.get_features_to_log()] + [label_name]
        features_to_log = set(features_to_log)
        for batch_count, (x, y) in enumerate(test_dataset.take(-1)):  # returns (x, y) tuples
            if batch_count:
                for key in x.keys():
                    if key in features_to_log:  # only log if necessary
                        # Here I am appending, np.vstack or other numpy tricks will be slow
                        # as they create new arrays (a np.array require contiguous memory)
                        predictions[key].append(np.squeeze(x[key].numpy()))
                predictions[label_name].append(np.squeeze(y.numpy()))
            else:  # we initialize the {key: list()} in the 1st batch
                for key in x.keys():
                    if key in features_to_log:
                        predictions[key] = [np.squeeze(x[key].numpy())]
                predictions[label_name] = [np.squeeze(y.numpy())]
            if batch_count % logging_frequency == 0:
                self.logger.info(f"Finished evaluating {batch_count} batches")
        # This is a memory bottleneck; we bring everything in memory
        for key, val in predictions.items():
            # we want to create np.arrays that contain the full data here.
            # to do this, if something is 2-dim we use np.vstack
            # for 1-dim data we use np.hstack for the intended result
            predictions[key] = np.hstack(val) if len(predictions[key][0].shape) == 1 else np.vstack(
                val)
        predictions_df = pd.DataFrame({key: val if len(val.shape) == 1 else [inner for inner in val]
                                       for key, val in predictions.items()})
        return predictions_df



import yaml
from logging import Logger

from ml4ir.base.features.feature_config import SequenceExampleFeatureConfig
import ml4ir.base.io.file_io as file_io

from typing import Optional


class RankingFeatureConfig(SequenceExampleFeatureConfig):
    def extract_features(self, features_dict, logger: Optional[Logger] = None):
        try:
            self.rank = features_dict.get("rank")
            self.all_features.append(self.rank)
        except KeyError:
            self.rank = None
            if logger:
                logger.warning("'rank' key not found in the feature_config specified")


def parse_config(feature_config, logger: Optional[Logger] = None) -> RankingFeatureConfig:
    if feature_config.endswith(".yaml"):
        feature_config = file_io.read_yaml(feature_config)
        if logger:
            logger.info("Reading feature config from YAML file : {}".format(feature_config))
    else:
        feature_config = yaml.safe_load(feature_config)
        if logger:
            logger.info("Reading feature config from YAML string : {}".format(feature_config))

    return SequenceExampleFeatureConfig(feature_config, logger=logger)



from ml4ir.applications.classification.model.metrics.metrics_impl import Top5CategoricalAccuracy


def test_top_5_categorical_accuracy():
    """Test for calculating top_5_categorical_accuracy metric"""
    metric = Top5CategoricalAccuracy(feature_config=None, metadata_features={})
    metric.update_state(
        [[[0, 0, 1, 0, 0, 0]], [[0, 1, 0, 0, 0, 0]]],
        [[[0.1, 0.1, 0.4, 0.2, 0.1, 0.1]], [[0.19, 0.01, 0.4, 0.2, 0.1, 0.1]]],
    )
    score = metric.result().numpy()
    assert score == 0.5



import unittest
import tensorflow as tf
import numpy as np
import os
import random
import gc
from argparse import Namespace
from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer

from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.model.scoring.scoring_model import ScorerBase, RelevanceScorer
from ml4ir.base.model.scoring.interaction_model import InteractionModel, UnivariateInteractionModel
from ml4ir.base.model.optimizers.optimizer import get_optimizer
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.io.logging_utils import setup_logging
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import ArchitectureKey
from ml4ir.base.tests.test_base import RelevanceTestBase
from ml4ir.applications.ranking.model.ranking_model import RankingModel, LinearRankingModel
from ml4ir.applications.ranking.model.losses import loss_factory
from ml4ir.applications.ranking.model.metrics import metric_factory
from ml4ir.applications.ranking.config.parse_args import get_args

import warnings
from typing import Union, List, Type

warnings.filterwarnings("ignore")


OUTPUT_DIR = "ml4ir/applications/ranking/tests/test_output"
ROOT_DATA_DIR = "ml4ir/applications/ranking/tests/data"
FEATURE_CONFIG_FNAME = "feature_config.yaml"


class RankingTestBase(RelevanceTestBase):

    def get_ranking_model(
        self,
        loss_key: str,
        metrics_keys: List,
        feature_config: FeatureConfig,
        model_config: dict = {},
        feature_layer_keys_to_fns={},
        initialize_layers_dict={},
        freeze_layers_list=[],
    ) -> RelevanceModel:
        """
        Creates RankingModel

        NOTE: Override this method to create custom loss, scorer, model objects
        """

        # Define interaction model
        interaction_model: InteractionModel = UnivariateInteractionModel(
            feature_config=feature_config,
            feature_layer_keys_to_fns=feature_layer_keys_to_fns,
            tfrecord_type=self.args.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            file_io=self.file_io,
        )

        # Define loss object from loss key
        loss: RelevanceLossBase = loss_factory.get_loss(
            loss_key=loss_key, scoring_type=self.args.scoring_type
        )

        # Define scorer
        scorer: ScorerBase = RelevanceScorer(
            feature_config=feature_config,
            model_config=self.model_config,
            interaction_model=interaction_model,
            loss=loss,
            output_name=self.args.output_name,
            logger=self.logger,
            file_io=self.file_io,
        )

        # Define metrics objects from metrics keys
        metrics: List[Union[Type[Metric], str]] = [
            metric_factory.get_metric(metric_key=metric_key) for metric_key in metrics_keys
        ]

        # Define optimizer
        optimizer: Optimizer = get_optimizer(
            model_config=self.file_io.read_yaml(self.args.model_config),
        )

        # Combine the above to define a RelevanceModel
        if self.model_config["architecture_key"] == ArchitectureKey.LINEAR:
            RankingModelClass = LinearRankingModel
        else:
            RankingModelClass = RankingModel
        relevance_model: RelevanceModel = RankingModelClass(
            feature_config=feature_config,
            tfrecord_type=self.args.tfrecord_type,
            scorer=scorer,
            metrics=metrics,
            optimizer=optimizer,
            model_file=self.args.model_file,
            initialize_layers_dict=initialize_layers_dict,
            freeze_layers_list=freeze_layers_list,
            compile_keras_model=self.args.compile_keras_model,
            output_name=self.args.output_name,
            logger=self.logger,
            file_io=self.file_io,
        )

        return relevance_model


if __name__ == "__main__":
    unittest.main()



import os
import numpy as np

from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.applications.ranking.model.ranking_model import RankingModel
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.applications.ranking.tests.test_base import RankingTestBase

# Constants
GOLD_METRICS = {'query_count': 1500.0,
                'old_ACR': 1.656,
                'new_ACR': 2.410,
                'old_MRR': 0.783,
                'new_MRR': 0.597,
                'old_name_match_failure_all_mean': 0.061,
                'old_name_match_failure_any_mean': 0.111,
                'old_name_match_failure_all_rank_mean': 0.142,
                'old_name_match_failure_any_rank_mean': 0.351,
                'old_name_match_failure_any_count_mean': 0.165,
                'old_name_match_failure_any_fraction_mean': 0.087,
                'new_name_match_failure_all_mean': 0.086,
                'new_name_match_failure_any_mean': 0.218,
                'new_name_match_failure_all_rank_mean': 0.196,
                'new_name_match_failure_any_rank_mean': 0.788,
                'new_name_match_failure_any_count_mean': 0.342,
                'new_name_match_failure_any_fraction_mean': 0.153,
                'perc_improv_ACR': -45.513,
                'perc_improv_MRR': -23.760,
                'perc_improv_name_match_failure_all_mean': -40.217,
                'perc_improv_name_match_failure_any_mean': -96.407,
                'perc_improv_name_match_failure_all_rank_mean': -38.028,
                'perc_improv_name_match_failure_any_rank_mean': -124.478,
                'perc_improv_name_match_failure_any_count_mean': -106.854,
                'perc_improv_name_match_failure_any_fraction_mean': -75.401}


class RankingModelTest(RankingTestBase):
    def run_default_pipeline(self, data_dir: str, data_format: str, feature_config_path: str):
        """Train a model with the default set of args"""
        feature_config: FeatureConfig = FeatureConfig.get_instance(
            tfrecord_type=self.args.tfrecord_type,
            feature_config_dict=self.file_io.read_yaml(feature_config_path),
            logger=self.logger,
        )
        data_dir = os.path.join(self.root_data_dir, "tfrecord")
        data_format = "tfrecord"

        metrics_keys = ["categorical_accuracy", "MRR", "ACR"]

        relevance_dataset = RelevanceDataset(
            data_dir=data_dir,
            data_format=data_format,
            feature_config=feature_config,
            tfrecord_type=self.args.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns={},
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.file_io,
            logger=self.logger,
        )

        ranking_model: RankingModel = self.get_ranking_model(
            loss_key=self.args.loss_key, feature_config=feature_config, metrics_keys=metrics_keys
        )

        overall_metrics, _, _ = ranking_model.evaluate(
            test_dataset=relevance_dataset.test, logs_dir=self.args.logs_dir,
        )

        return overall_metrics.to_dict()

    def test_model_training(self):
        """
        Test model training and evaluate the performance metrics
        """

        # Test model training on TFRecord SequenceExample data
        data_dir = os.path.join(self.root_data_dir, "tfrecord")
        feature_config_path = os.path.join(self.root_data_dir, "configs", self.feature_config_fname)

        metrics = self.run_default_pipeline(
            data_dir=data_dir, data_format="tfrecord", feature_config_path=feature_config_path
        )

        # Compare the metrics to gold metrics
        for gold_metric_name, gold_metric_val in GOLD_METRICS.items():
            assert gold_metric_name in metrics
            assert np.isclose(metrics[gold_metric_name], gold_metric_val, atol=0.02)



# Generating csv test data for classification model in tests.

import string
import csv

from random import seed
from random import randint


CSV_TRAIN_FILE_PATH = "data/csv/train/file_0.csv"
CSV_TEST_FILE_PATH = "data/csv/test/file_0.csv"
CSV_VALIDATION_FILE_PATH = "data/csv/validation/file_0.csv"
CSV_VOCABULARY_QUERY_FILE_PATH = "data/configs/vocabulary/query_word.csv"
CSV_VOCABULARY_DOMAIN_ID_FILE_PATH = "data/configs/vocabulary/domain_id.csv"
CSV_VOCABULARY_ENTITY_FILE_PATH = "data/configs/vocabulary/entity_id.csv"

CSV_DATA_COLUMNS_HEADER = ["query_key", "query_text", "domain_id", "user_context", "code", "entity_id"]
CSV_VOCABULARY_QUERY_COLUMNS_HEADER = ["word"]
CSV_VOCABULARY_DOMAIN_ID_COLUMNS_HEADER = ["domain_id"]
CSV_VOCABULARY_ENTITY_COLUMNS_HEADER = ["entity"]

VOCABULARY_QUERY = [
    "the", "tragedy", "of", "hamlet", "prince", "denmark", "shakespeare", "homepage", "entire", "play", "act", "i",
    "scene", "elsinore", "a", "platform", "before", "castle", "francisco", "at", "his", "post", "enter", "to", "him",
    "bernardo", "whos", "there", "nay", "answer", "me", "stand", "and", "unfold", "yourself", "long", "live", "king",
    "he", "you", "come", "most", "carefully", "upon", "your", "hour", "tis", "now", "struck", "twelve", "get", "thee",
    "bed", "for", "this", "relief", "much", "thanks", "bitter", "cold", "am", "sick", "heart", "have", "had", "quiet",
    "guard", "not", "mouse", "stirring", "well", "good", "night", "if", "do", "meet", "horatio", "marcellus", "rivals",
    "my", "watch", "bid", "them", "make", "haste", "think", "hear", "ho", "friends", "ground", "liegemen", "dane"
]
VOCABULARY_FEATURE_DOMAIN_ID = [char for char in string.ascii_lowercase.upper()]
VOCABULARY_FEATURE_ENTITY = ["AAA", "BBB", "CCC", "DDD", "EEE", "FFF", "GGG", "HHH"]
# toy feature (untrainable); used for grouped metric calculation:
VOCABULARY_FEATURE_CODE = ["local", "production", "canary"]
VOCABULARY_LABEL = VOCABULARY_FEATURE_ENTITY[:5]

VOCABULARY_PATH_HEADER_DATA = [
    (CSV_VOCABULARY_QUERY_FILE_PATH, CSV_VOCABULARY_QUERY_COLUMNS_HEADER, VOCABULARY_QUERY),
    (CSV_VOCABULARY_DOMAIN_ID_FILE_PATH, CSV_VOCABULARY_DOMAIN_ID_COLUMNS_HEADER, VOCABULARY_FEATURE_DOMAIN_ID),
    (CSV_VOCABULARY_ENTITY_FILE_PATH, CSV_VOCABULARY_ENTITY_COLUMNS_HEADER, VOCABULARY_FEATURE_ENTITY)
]

TOTAL_DATA_SIZE = 1000
PARTITION_TRAIN = 0.7
PARTITION_VALIDATION = 0.1
PARTITION_TEST = 0.2


class FeatureGenerator:
    """Helper to randomly generate features based on a given vocabulary list."""
    def __init__(self, vocabulary, max_token_length, sequence_joiner=" "):
        self.__max_token_length = max_token_length
        self.__vocabulary = vocabulary
        self.__vocabulary_length = len(vocabulary) - 1
        self.__sequence_joiner = sequence_joiner

    def generate_feature(self):
        query_tokens = [self.__vocabulary[randint(0, self.__vocabulary_length)]
                        for _ in range(0, randint(1, self.__max_token_length))]
        return self.__sequence_joiner.join(query_tokens)


def generate_csv_test_data():
    """Generates data under classification/tests/csv folder."""
    seed(123)
    feature_query_text_generator = FeatureGenerator(VOCABULARY_QUERY, 7)
    feature_domain_id_generator = FeatureGenerator(VOCABULARY_FEATURE_DOMAIN_ID, 1)
    feature_user_context_generator = FeatureGenerator(VOCABULARY_FEATURE_ENTITY, 20, sequence_joiner=",")
    feature_code_generator = FeatureGenerator(VOCABULARY_FEATURE_CODE, 1)
    label_generator = FeatureGenerator(VOCABULARY_LABEL, 1)
    generators = [
        feature_query_text_generator,
        feature_domain_id_generator,
        feature_user_context_generator,
        feature_code_generator,
        label_generator
    ]

    for (path, number_rows) in [(CSV_TRAIN_FILE_PATH, TOTAL_DATA_SIZE * PARTITION_TRAIN),
                                (CSV_TEST_FILE_PATH, TOTAL_DATA_SIZE * PARTITION_TEST),
                                (CSV_VALIDATION_FILE_PATH, TOTAL_DATA_SIZE * PARTITION_VALIDATION)]:
        rows = [CSV_DATA_COLUMNS_HEADER] + [['query_id_' + str(idx)] + [g.generate_feature() for g in generators]
                                   for idx in range(0, int(number_rows))]
        with open(path, 'w', newline='') as file:
            writer = csv.writer(file, quoting=csv.QUOTE_NONNUMERIC)
            writer.writerows(rows)


def generate_csv_vocabulary_data():
    """Generates vocabulary data under classification/tests/csv folder."""
    for (path, header, data) in VOCABULARY_PATH_HEADER_DATA:
        rows = [header] + [[word] for word in data]
        with open(path, 'w', newline='') as file:
            writer = csv.writer(file, quoting=csv.QUOTE_NONNUMERIC)
            writer.writerows(rows)


def main():
    generate_csv_test_data()
    generate_csv_vocabulary_data()
    return


if __name__ == "__main__":
    """
    Generates data under classification/tests folder used
     to train and assess classification model in tests.
    """
    main()



import ast
from typing import List, Dict


DOT = "."
NAME = "name"


def cast_dynamic_val(val: str):
    """
    Cast the dynamic val from string to python base data type

    Parameters
    ----------
    val: str
        String value to be cast

    Returns
    -------
    Value cast to appropriate dtype
    """
    if isinstance(val, str):
        if val.replace(DOT, "").isdigit():
            if DOT in val:
                # float
                return float(val)
            else:
                # int
                return int(val)
        elif val.isalnum():
            # bool
            if val.lower() == "true":
                return True
            elif val.lower() == "false":
                return False
            # str
            return val
        else:
            # If contains punctuation
            return ast.literal_eval(val)
    else:
        return val


def override_list(base_list: List, dynamic_key: str, val):
    """
    Customize the base list by updating with the
    dynamic_key and val.

    Parameters
    ----------
    base: dict
        Dictionary or List to be customized with dynamic args
    dynamic_key: str
        Key to identify the location the value should be updated.
        Nested with DOT like "custom.key_0.key_1.key_2.0.0.key_4"
    val: str or float or int or dict or list
        Value to be set

    Returns
    -------
    dict
        Updated base_list based on the key-value pairs in dynamic_args

    Notes
    -----
    This will be called recursively with override_dict.
    If dynamic_key is not a number, then we try to match on `name` field
    in the list of dictionaries.
    """
    def find_root_key_index(base_list, root_key):
        if root_key.isdigit():
            # If array index
            root_key = int(root_key)
        else:
            # If string, then match on `name`
            for root_key_i in range(len(base_list)):
                if root_key == base_list[root_key_i][NAME]:
                    root_key = root_key_i
                    break

            if not isinstance(root_key, int):
                raise KeyError("{} not found in List".format(root_key))

        return root_key

    if DOT in dynamic_key:
        # Compute root and subtree keys
        root_key = find_root_key_index(base_list, dynamic_key.split(DOT)[0])
        subtree_key = DOT.join(dynamic_key.split(DOT)[1:])

        # Extract subtree
        subtree = base_list[root_key]

        if isinstance(subtree, dict):
            root_val = override_dict(base_dict=subtree,
                                     dynamic_key=subtree_key,
                                     val=val)
        elif isinstance(subtree, list):
            root_val = override_list(base_list=subtree,
                                     dynamic_key=subtree_key,
                                     val=val)
        else:
            raise ValueError(
                "Unsupported subtree type. Must be one of list or dict")
    else:
        # End of nested dynamic key
        root_key = find_root_key_index(base_list, dynamic_key)
        root_val = val

    base_list[root_key] = root_val

    return base_list


def override_dict(base_dict: Dict, dynamic_key: str, val):
    """
    Customize the base dictionary by updating with the
    dynamic_key and val.

    Parameters
    ----------
    base: dict
        Dictionary or List to be customized with dynamic args
    dynamic_key: str
        Key to identify the location the value should be updated.
        Nested with DOT like "custom.key_0.key_1.key_2.0.0.key_4"
    val: str or float or int or dict or list
        Value to be set

    Returns
    -------
    dict
        Updated base_dict based on the key-value pairs in dynamic_args

    Notes
    -----
    This will be called recursively along with override_list
    """
    if DOT in dynamic_key:
        # Compute root and subtree keys
        root_key = dynamic_key.split(DOT)[0]
        subtree_key = DOT.join(dynamic_key.split(DOT)[1:])

        # Extract subtree
        subtree = base_dict[root_key]

        if isinstance(subtree, dict):
            root_val = override_dict(base_dict=subtree,
                                     dynamic_key=subtree_key,
                                     val=val)
        elif isinstance(subtree, list):
            root_val = override_list(base_list=subtree,
                                     dynamic_key=subtree_key,
                                     val=val)
        else:
            raise ValueError(
                "Unsupported subtree type. Must be one of list or dict")
    else:
        # End of nested dynamic key
        root_key = dynamic_key
        root_val = val

    base_dict[root_key] = root_val

    return base_dict


def override_with_dynamic_args(base_dict: Dict, dynamic_args: Dict = {}):
    """
    Customize the base dictionary by updating with the
    key-value pairs in dynamic args.

    Parameters
    ----------
    base_dict: dict
        Dictionary to be customized with dynamic args
    dynamic_args: dict
        Flat dictionary with keys and values used to update the base_dict

    Returns
    -------
    dict
        Updated base_dict based on the key-value pairs in dynamic_args

    Notes
    -----
    This is used to override
    base feature_config and model_config presets from command line arguments.
    """
    for key, val in dynamic_args.items():
        base_dict = override_dict(base_dict=base_dict,
                                  dynamic_key=key,
                                  val=cast_dynamic_val(val))

    return base_dict



import unittest
import warnings
import tensorflow as tf
from ml4ir.base.model.architectures.fixed_additive_positional_bias import FixedAdditivePositionalBias

warnings.filterwarnings("ignore")

INPUT_DIR = "/ml4ir/applications/ranking/tests/data/configs"

class TestFixedAdditivePositionalBias(unittest.TestCase):

    def apply_additive_positional_bias(self, rank_index, max_ranks, training):
        """Tests if the positional bias is applied as expected during training (training=True) or evaluation."""
        positional_bias = FixedAdditivePositionalBias(max_ranks)
        biases = positional_bias(tf.convert_to_tensor(rank_index), training)
        if not training:
            assert tf.math.reduce_sum(biases).numpy() == 0.0
        else:
            for i in range(len(rank_index)):
                assert biases[i][0].numpy() == positional_bias.weights[0][rank_index[i] - 1].numpy()[0]

    def test_one_hot_conversion(self):
        """calling additive positional bias"""
        self.apply_additive_positional_bias([5, 4], 5, True)
        self.apply_additive_positional_bias([1,3], 5, False)
        self.apply_additive_positional_bias([1,2], 2, True)
        self.apply_additive_positional_bias([2], 2, False)
        self.apply_additive_positional_bias([2,4,6,8], 10, True)

    def test_zeros_weight_initialization(self):
        """Testing weight initializations to zeros"""
        positional_bias = FixedAdditivePositionalBias(max_ranks=5, kernel_initializer='Zeros')
        positional_bias(tf.constant([1.]))
        weights = positional_bias.dense.get_weights()
        assert all([w == 0. for w in weights[0]])

    def test_non_zeros_weight_initialization(self):
        """Testing weight initializations to non zeros"""
        positional_bias = FixedAdditivePositionalBias(max_ranks=5, kernel_initializer='glorot_uniform')
        positional_bias(tf.constant([1.]))
        weights = positional_bias.dense.get_weights()
        assert any([w != 0. for w in weights[0]])



if __name__ == "__main__":
    unittest.main()



import tensorflow as tf
from tensorflow.keras import layers
from ml4ir.applications.ranking.config.keys import PositionalBiasHandler
from tensorflow.keras import regularizers


class FixedAdditivePositionalBias(layers.Layer):
    """
    This class implements a solution to handle positional bias in logged data.
    During training, a bias term is learned for each rank and is added to the relevance score of each query-document
    pair depending on its rank.
    The process of adding the positional bias is only applied during training because this is a way to de-bias data
    from logs (typically clicked/no clicked pairs). During inference we do not know the actual ranks, so the addition
    is turned off.
    This is an approach we have found useful for logged data as these are presented to the user with some order and
    there is a presentation bias (that this technique tries to model).
    When the data is labeled with actual graded relevance annotations then this technique is not recommended.

    To trigger this technique add this section to the model_config.yaml file:

    positional_bias_handler:
        key: fixed_additive_positional_bias
        max_ranks_count: x

    Where x is the maximum number of documents per query.
    """
    def __init__(self, max_ranks, kernel_initializer='Zeros', l1_coeff=0, l2_coeff=0):
        super(FixedAdditivePositionalBias, self).__init__()
        self.dense = layers.Dense(1,
                     name=PositionalBiasHandler.FIXED_ADDITIVE_POSITIONAL_BIAS,
                     kernel_initializer=kernel_initializer,
                     kernel_regularizer=regularizers.l1_l2(l1=l1_coeff, l2=l2_coeff),
                     activation=None,
                     use_bias=False)
        self.max_ranks = max_ranks

    def call(self, inputs, training=False):
        """
        Invoke the positional bias handling

        Parameters
        ----------
        input : Tensor object
            rank index tensor
        training : bool
            If the layer should be run as training or not

        Returns
        -------
        Tensor object
            positional biases resulting from a feedforwrd of the converted one hot tensor through a dense layer.
        """
        features = tf.one_hot(tf.cast(tf.subtract(inputs, 1), dtype=tf.int64), depth=self.max_ranks,
                              dtype=tf.dtypes.float32)
        if not training:
            features = tf.multiply(features, 0.0)
        return self.dense(features)



import unittest
import ml4ir.base.config.dynamic_args as dynamic_args
from ml4ir.base.config.parse_args import RelevanceArgParser
from ml4ir.base.io.local_io import LocalIO

FEATURE_CONFIG_PATH = "ml4ir/applications/ranking/tests/data/configs/feature_config.yaml"
MODEL_CONFIG_PATH = "ml4ir/base/config/default_model_config.yaml"


class DynamicArgsTest(unittest.TestCase):

    def setUp(self):
        file_io = LocalIO()
        self.feature_config_dict = file_io.read_yaml(FEATURE_CONFIG_PATH)
        self.model_config_dict = file_io.read_yaml(MODEL_CONFIG_PATH)

    def test_dynamic_args(self):
        """
        Unit test the accepted dynamic args by argparse
        """
        parser = RelevanceArgParser()

        # Test empty dynamic args
        args = parser.parse_args([])
        assert "feature_config_custom" not in args
        assert "model_config_custom" not in args

        # Test feature config and model config custom args
        args = parser.parse_args([
            "--feature_config.features.query_text.feature_layer_info.args.encoding_size",
            "512",
            "--model_config.layers.first_dense.units",
            "1024"
        ])
        assert len(args.feature_config_custom) == 1
        assert len(args.model_config_custom) == 1
        assert args.feature_config_custom[
            "features.query_text.feature_layer_info.args.encoding_size"] == "512"
        assert args.model_config_custom[
            "layers.first_dense.units"] == "1024"

    def test_cast_dynamic_val(self):
        """
        Unit test the value casting method
        """
        # Test string casting
        assert isinstance(dynamic_args.cast_dynamic_val("abc"), str)
        assert isinstance(dynamic_args.cast_dynamic_val("123"), int)
        assert isinstance(dynamic_args.cast_dynamic_val("12.3"), float)
        assert isinstance(dynamic_args.cast_dynamic_val("[1, 2, 3]"), list)
        assert isinstance(dynamic_args.cast_dynamic_val(
            "{'a': 1, 'b': 2}"), dict)

        # Test no casting when input val is not string
        assert isinstance(dynamic_args.cast_dynamic_val(123), int)
        assert isinstance(dynamic_args.cast_dynamic_val(12.3), float)
        assert isinstance(dynamic_args.cast_dynamic_val([1, 2, 3]), list)
        assert isinstance(
            dynamic_args.cast_dynamic_val({'a': 1, 'b': 2}), dict)

    def test_override_with_dynamic_args(self):
        """
        Unit test the override_with_dynamic_args method

        Check if the dictionary of dynamic arguments are applied to the
        base dictionary
        """

        # Test feature config updates
        base_dict = self.feature_config_dict
        assert not base_dict["features"][4]["feature_layer_info"]["args"]["encoding_size"] == 1024
        assert not base_dict["features"][6]["feature_layer_info"]["args"]["embedding_size"] == 1024

        base_dict = dynamic_args.override_with_dynamic_args(
            base_dict,
            {
                "features.4.feature_layer_info.args.encoding_size": "1024",
                "features.domain_name.feature_layer_info.args.embedding_size": 1024
            })

        assert base_dict["features"][4]["feature_layer_info"]["args"]["encoding_size"] == 1024
        assert base_dict["features"][6]["feature_layer_info"]["args"]["embedding_size"] == 1024

        # Test model config updates
        base_dict = self.model_config_dict
        assert not base_dict["optimizer"]["key"] == "sgd"
        assert not base_dict["lr_schedule"]["learning_rate"] == 0.00005

        base_dict = dynamic_args.override_with_dynamic_args(
            base_dict,
            {
                "optimizer.key": "sgd",
                "lr_schedule.learning_rate": "0.00005"
            })

        assert base_dict["optimizer"]["key"] == "sgd"
        assert base_dict["lr_schedule"]["learning_rate"] == 0.00005

    def test_override_dict(self):
        """
        Unit test the override_dict method

        Check if a dictionary can be overriden with a value
        """
        # Test high level key
        assert dynamic_args.override_dict(
            {
                "key_0": "val_0",
                "key_1": "val_1"
            },
            "key_0",
            "val_x")["key_0"] == "val_x"

        # Test nested dict
        assert dynamic_args.override_dict(
            {
                "key_0": {
                    "key_0_0": "val_0_0",
                    "key_0_1": "val_0_1"
                },
                "key_1": "val_1"
            },
            "key_0.key_0_1",
            "val_x")["key_0"]["key_0_1"] == "val_x"

    def test_override_list(self):
        """
        Unit test the override_dict method

        Check if a list can be overriden with a value
        """
        assert dynamic_args.override_list(
            [
                "val_0",
                "val_1",
                "val_2"
            ],
            "2",
            "val_x"
        )[2] == "val_x"

        # Test updating by matching on `name`
        assert dynamic_args.override_list(
            [
                {"name": "val_0"},
                {"name": "val_1"}
            ],
            "val_0.name",
            "val_x"
        )[0]["name"] == "val_x"

        # Test updating by array index
        assert dynamic_args.override_list(
            [
                {"name": "val_0"},
                {"name": "val_1"}
            ],
            "0.name",
            "val_x"
        )[0]["name"] == "val_x"

    def test_new_arg(self):
        """
        Unit test adding a new key value argument
        """
        updated_dict = dynamic_args.override_dict(
            {
                "key_0": "val_0",
                "key_1": "val_1"
            },
            "key_2",
            "val_2")

        assert len(updated_dict) == 3
        assert updated_dict["key_2"] == "val_2"



import ast
from argparse import ArgumentParser, Namespace, Action
from typing import List
from ml4ir.base.config.keys import OptimizerKey, DataFormatKey, TFRecordTypeKey, ExecutionModeKey, ServingSignatureKey, FileHandlerKey
from ml4ir.applications.ranking.config.keys import LossKey as RankingLoss, MetricKey as RankingMetricKey
from ml4ir.applications.classification.config.keys import LossKey as ClassificationLoss, MetricKey as ClassificationMetricKey


class CustomFeatureDictUpdater(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        """
        Add a value to a custom argument dictionary.
        If dictionary does not exist a new one is created.
        If dictionary already exists, it is updated with new key-value pair.

        Can be used to collect all custom arguments for feature_config
        by specifying as feature_config.custom_arg 128. This will update the
        dictionary for feature_config custom args with the new key "custom_arg"
        and corresponding value "128".

        Parameters
        ----------
        parser: ArgumentParser
            ArgumentParser to use this action with
        namespace: Namespace
            The namespace object that will be updated with the parsed value
        values: str
            Value to be added to dictionary
        option_string: str
            Name of command line argument the action is used with

        Returns
        -------
        dict
            Dictionary with the updated key value pair
        """
        old_custom_args = getattr(namespace, self.dest)
        new_custom_args = {".".join(option_string.split(".")[1:]): values}
        if isinstance(old_custom_args, dict):
            new_custom_args.update(old_custom_args)
        setattr(namespace, self.dest, new_custom_args)


class RelevanceArgParser(ArgumentParser):
    """Defines the parser for the command line arguments for RelevancePipeline"""

    def __init__(self, allow_abbrev=False, **kwargs):
        super().__init__(allow_abbrev=allow_abbrev, **kwargs)
        self.define_args()
        self.set_default_args()


    def define_args(self):
        self.add_argument(
            "--data_dir",
            type=str,
            help="Path to the data directory to be used for training and inference. "
            "Can optionally include train/ val/ and test/ subdirectories. "
            "If subdirectories are not present, data will be split based on train_pcent_split",
        )

        self.add_argument(
            "--data_format",
            type=str,
            choices=DataFormatKey.get_all_keys(),
            default="tfrecord",
            help="Format of the data to be used. "
            "Supported Data Formats  are specified in ml4ir/base/config/keys.py",
        )

        self.add_argument(
            "--tfrecord_type",
            type=str,
            choices=TFRecordTypeKey.get_all_keys(),
            default="example",
            help="TFRecord type of the data to be used. "
            "Supported TFRecord type are specified in ml4ir/base/config/keys.py",
        )

        self.add_argument(
            "--feature_config",
            type=str,
            help="Path to YAML file or YAML string with feature metadata for training.",
        )

        self.add_argument(
            "--model_file",
            type=str,
            default=None,
            required=False,
            help="Path to a pretrained model to load for either resuming training or for running in"
            "inference mode.",
        )

        self.add_argument(
            "--model_config",
            type=str,
            default="ml4ir/base/config/default_model_config.yaml",
            help="Path to the Model config YAML used to build the model architecture.",
        )

        self.add_argument(
            "--loss_key",
            type=str,
            choices=RankingLoss.get_all_keys() + ClassificationLoss.get_all_keys(),
            help="Loss to optimize."
        )

        self.add_argument(
            "--metrics_keys",
            type=str,
            nargs="+",
            default=None,
            choices=RankingMetricKey.get_all_keys() + ClassificationMetricKey.get_all_keys(),
            help="A space separated list of metrics to compute.",
        )

        self.add_argument(
            "--monitor_metric",
            type=str,
            default=None,
            choices=RankingMetricKey.get_all_keys() + ClassificationMetricKey.get_all_keys(),
            help="Metric name to use for monitoring training loop in callbacks.",
        )

        self.add_argument(
            "--monitor_mode",
            type=str,
            default=None,
            help="Metric mode to use for monitoring training loop in callbacks",
        )

        self.add_argument(
            "--num_epochs",
            type=int,
            default=5,
            help="Max number of training epochs(or full pass over the data)",
        )

        self.add_argument(
            "--batch_size",
            type=int,
            default=128,
            help="Number of data samples to use per batch.",
        )

        self.add_argument(
            "--compute_intermediate_stats",
            type=ast.literal_eval,
            default=True,
            help="Whether to compute intermediate stats on test set (mrr, acr, etc) (slow)",
        )

        self.add_argument(
            "--execution_mode",
            type=str,
            choices=ExecutionModeKey.get_all_keys(),
            default="train_inference_evaluate",
            help="Execution mode for the pipeline.",
        )

        self.add_argument(
            "--random_state",
            type=int,
            default=123,
            help="Initialize the seed to control randomness for replication",
        )

        self.add_argument(
            "--run_id",
            type=str,
            default="",
            help="Unique string identifier for the current training run. "
                 "Used to identify logs and models directories. "
                 "Autogenerated if not specified.",
        )

        self.add_argument(
            "--run_group",
            type=str,
            default="general",
            help="Unique string identifier to group multiple model training runs."
                 " Allows for defining a meta grouping to filter different model "
                 "training runs for best model selection as a post step.",
        )

        self.add_argument(
            "--run_notes",
            type=str,
            default="",
            help="Notes for the current training run. Use this argument "
                 "to add short description of the model training run that "
                 "helps in identifying the run later.",
        )

        self.add_argument(
            "--models_dir",
            type=str,
            default="models/",
            help="Path to save the model. Will be expanded to models_dir/run_id",
        )

        self.add_argument(
            "--logs_dir",
            type=str,
            default="logs/",
            help="Path to save the training/inference logs. "
                 "Will be expanded to logs_dir/run_id",
        )

        self.add_argument(
            "--checkpoint_model",
            type=ast.literal_eval,
            default=True,
            help="Whether to save model checkpoints at the end of each epoch. Recommended - set to True",
        )

        self.add_argument(
            "--train_pcent_split",
            type=float,
            default=0.8,
            help="Percentage of all data to be used for training. The remaining is used for validation and "
            "testing. Remaining data is split in half if val_pcent_split or test_pcent_split are not "
            "specified.",
        )

        self.add_argument(
            "--val_pcent_split",
            type=float,
            default=-1,
            help="Percentage of all data to be used for testing.",
        )

        self.add_argument(
            "--test_pcent_split",
            type=float,
            default=-1,
            help="Percentage of all data to be used for testing.",
        )

        self.add_argument(
            "--max_sequence_size",
            type=int,
            default=0,
            help="Maximum number of elements per sequence feature.",
        )

        self.add_argument(
            "--inference_signature",
            type=str,
            choices=ServingSignatureKey.get_all_keys(),
            default="serving_default",
            help="SavedModel signature to be used for inference",
        )

        self.add_argument(
            "--use_part_files",
            type=ast.literal_eval,
            default=False,
            help="Whether to look for part files while loading data",
        )

        self.add_argument(
            "--logging_frequency",
            type=int,
            default=25,
            help="How often to log results to log file. Int representing number of batches.",
        )

        self.add_argument(
            "--group_metrics_min_queries",
            type=int,
            default=None,
            help="Minimum number of queries per group to be used to computed groupwise metrics.",
        )

        self.add_argument(
            "--compile_keras_model",
            type=ast.literal_eval,
            default=False,
            help="Whether to compile a loaded SavedModel into a Keras model. "
            "NOTE: This requires that the SavedModel's architecture, loss, metrics, etc are the same as the RankingModel"
            "If that is not the case, then you can still use a SavedModel from a model_file for inference/evaluation only",
        )

        self.add_argument(
            "--use_all_fields_at_inference",
            type=ast.literal_eval,
            default=False,
            help="Whether to require all fields in the serving signature of the SavedModel."
                 " If set to False, only requires fields with required_only=True",
        )

        self.add_argument(
            "--pad_sequence_at_inference",
            type=ast.literal_eval,
            default=False,
            help="Whether to pad sequence at inference time. "
                 "Used to define the TFRecord serving signature in the SavedModel",
        )

        self.add_argument(
            "--output_name",
            type=str,
            default="relevance_score",
            help="Name of the output node of the model",
        )

        self.add_argument(
            "--early_stopping_patience",
            type=int,
            default=2,
            help="How many epochs to wait before early stopping on metric degradation",
        )

        self.add_argument(
            "--file_handler",
            type=str,
            default="local",
            choices=FileHandlerKey.get_all_keys(),
            help="String specifying the file handler to be used.",
        )

        self.add_argument(
            "--initialize_layers_dict",
            type=str,
            default="{}",
            help="Dictionary of pretrained layers to be loaded."
            "The key is the name of the layer to be assigned the pretrained weights."
            "The value is the path to the pretrained weights.",
        )

        self.add_argument(
            "--freeze_layers_list",
            type=str,
            default="[]",
            help="List of layer names that are to be frozen instead of training."
            "Usually coupled with initialize_layers_dict to load pretrained weights and freeze them",
        )

        self.add_argument(
            "--non_zero_features_only",
            type=ast.literal_eval,
            default=False,
            help="[Ranklib format only] Only non zero features are stored.",
        )

        self.add_argument(
            "--keep_additional_info",
            type=ast.literal_eval,
            default=False,
            help="[Ranklib format only] Option to keep additional info "
                 "(All info after the '#' in the format [key = val]).",
        )

        self.add_argument(
            "--kfold",
            type=int,
            default=1,
            help="The number of folds for K-fold Cross Validation. Must be > 2 if testset is included in folds and > 1 otherwise.",
        )

        self.add_argument(
            "--include_testset_in_kfold",
            type=ast.literal_eval,
            default=False,
            help="Merge the testset with the training and validation sets and perform kfold on the merged dataset.",
        )

        self.add_argument(
            "--kfold_analysis_metrics",
            type=str,
            default=[],
            nargs='*',
            help="Metric to use for post Kfold CV analysis.",
        )

    def set_default_args(self):
        pass

    def parse_args(self, args: List[str]) -> Namespace:
        """
        Parse command line arguments passed as a list of strings.
        Additionally, handles dynamic arguments for feature_config
        and model_config with prefixes feature_config. and model_config.
        respectively

        Parameters
        ----------
        args: list of str
            List of command line args to be parsed

        Returns
        -------
        Namespace
            Namespace object obtained by parsing the input
        """
        dynamic_args = self.parse_known_args(args)[1]

        for i in range(int(len(dynamic_args) / 2)):
            key = dynamic_args[i * 2]
            if key.split(".")[0] not in {"--feature_config", "--model_config"}:
                raise KeyError(
                    "Dynamic arguments currently supported must have the prefix feature_config. or model_config., but found: {}".format(key))
            dest = "{}_custom".format(key.split(".")[0]).replace("--", "")
            self.add_argument(key,
                              dest=dest,
                              action=CustomFeatureDictUpdater)

        return super(RelevanceArgParser, self).parse_args(args)


def get_args(args: List[str]) -> Namespace:
    return RelevanceArgParser().parse_args(args)



import yaml
from ml4ir.applications.classification.tests.test_base import ClassificationTestBase
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.model.architectures.dnn import DNN


class DNNModelTest(ClassificationTestBase):

    def test_num_classes_from_units(self):
        feature_config = None

        model_info = yaml.safe_load('''
            architecture_key: dnn
            layers:
              - type: dense
                name: first_dense
                units: 256
                activation: relu
              - type: dense
                name: final_dense
                units: 9
                activation: null
        ''')

        dnn = DNN(model_info, feature_config, self.file_io)
        assert(len(dnn.layer_ops) == 2)
        assert(dnn.layer_ops[0].get_config()['units'] == 256)
        assert(dnn.layer_ops[-1].get_config()['units'] == 9)

    def test_num_classes_from_vocabulary_file(self):
        feature_config = FeatureConfig(yaml.safe_load('''
            query_key:
              name: query_key
              node_name: query_key
              trainable: false
              dtype: string
            label:
              name: entity_id
              feature_layer_info:
                type: numeric
                fn: categorical_indicator_with_vocabulary_file
                args:
                  vocabulary_file: ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv
            features:
              - name: query_text
                trainable: false
                dtype: string
        '''))

        model_info = yaml.safe_load('''
            architecture_key: dnn
            layers:
              - type: dense
                name: first_dense
                units: 256
                activation: relu
              - type: dense
                name: final_dense
                activation: null
        ''')

        dnn = DNN(model_info, feature_config, self.file_io)
        assert(len(dnn.layer_ops) == 2)
        assert(dnn.layer_ops[0].get_config()['units'] == 256)
        assert(dnn.layer_ops[-1].get_config()['units'] == 9)

    def test_drop_out_layers(self):
        feature_config = FeatureConfig(yaml.safe_load('''
            query_key:
              name: query_key
              node_name: query_key
              trainable: false
              dtype: string
            label:
              name: entity_id
              feature_layer_info:
                type: numeric
                fn: categorical_indicator_with_vocabulary_file
                args:
                  vocabulary_file: ml4ir/applications/classification/tests/data/configs/vocabulary/entity_id.csv
            features:
              - name: query_text
                trainable: false
                dtype: string
        '''))

        model_info = yaml.safe_load('''
            architecture_key: dnn
            layers:
              - type: dense
                name: first_dense
                units: 256
                activation: relu
              - type: dropout
                name: first_dropout
                rate: 0.3
              - type: dense
                name: second_dense
                units: 64
                activation: relu
              - type: dropout
                name: second_dropout
                rate: 0.0
              - type: dense
                name: final_dense
                activation: null
        ''')

        dnn = DNN(model_info, feature_config, self.file_io)
        assert(len(dnn.layer_ops) == 5)
        assert(dnn.layer_ops[0].get_config()['units'] == 256)
        assert(dnn.layer_ops[1].get_config()['rate'] == 0.3)
        assert(dnn.layer_ops[2].get_config()['units'] == 64)
        assert(dnn.layer_ops[3].get_config()['rate'] == 0.0)
        assert(dnn.layer_ops[4].get_config()['units'] == 9)



from typing import Optional
from logging import Logger
import tensorflow as tf

from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.io.file_io import FileIO


class KfoldRelevanceDataset(RelevanceDataset):
    def __init__(
            self,
            data_dir: str,
            data_format: str,
            feature_config: FeatureConfig,
            tfrecord_type: str,
            file_io: FileIO,
            max_sequence_size: int = 0,
            batch_size: int = 128,
            preprocessing_keys_to_fns: dict = {},
            train_pcent_split: float = 0.8,
            val_pcent_split: float = -1,
            test_pcent_split: float = -1,
            use_part_files: bool = False,
            parse_tfrecord: bool = True,
            logger: Optional[Logger] = None,
            keep_additional_info: int = 0,
            non_zero_features_only: int = 0,
            num_folds: int = 3,
            include_testset_in_kfold: bool = False,
            read_data_sets: bool = False
    ):
        self.include_testset_in_kfold = include_testset_in_kfold
        self.num_folds = num_folds

        self.feature_config = feature_config
        self.max_sequence_size = max_sequence_size
        self.logger = logger
        self.data_dir = data_dir
        self.data_format: str = data_format
        self.tfrecord_type = tfrecord_type
        self.batch_size: int = batch_size
        self.preprocessing_keys_to_fns = preprocessing_keys_to_fns
        self.file_io = file_io

        self.train_pcent_split: float = train_pcent_split
        self.val_pcent_split: float = val_pcent_split
        self.test_pcent_split: float = test_pcent_split
        self.use_part_files: bool = use_part_files

        self.keep_additional_info = keep_additional_info
        self.non_zero_features_only = non_zero_features_only

        self.train: Optional[tf.data.TFRecordDataset] = None
        self.validation: Optional[tf.data.TFRecordDataset] = None
        self.test: Optional[tf.data.TFRecordDataset] = None
        if read_data_sets:
            self.create_dataset(parse_tfrecord)

    def merge_datasets(self):
        """
        Concat the datasets (training, validation, test) together

        Returns
        -------
        all_data: Tensorflow Dataset
            The final concatenated dataset.
        """

        if self.include_testset_in_kfold:
            all_data = self.train.concatenate(self.validation).concatenate(
                self.test)
        else:
            all_data = self.train.concatenate(self.validation)

        # un-batch and shuffle all queries
        all_data = all_data.unbatch()
        # shuffling before using the shard method gives unexpected results. Should be avoided
        # all_data = all_data.shuffle(batch_size * 2)
        return all_data

    def create_folds(self, fold_id, merged_data, relevance_dataset):
        """
        Create training, validation and test set according to the passed fold id.

        Parameters
        ----------
        fold_id : int
            current fold number
        merged_data: Tensorflow Dataset
            the dataset used to create folds
        relevance_dataset: RelevanceDataset object
            Used to access the test set to setup folds
        """
        test = None
        training_idx = list(range(self.num_folds))
        if self.include_testset_in_kfold:
            validation = merged_data.shard(self.num_folds, fold_id)
            test_idx = fold_id + 1
            if fold_id + 1 >= self.num_folds:
                test_idx = 0
            test = merged_data.shard(self.num_folds, test_idx)
            training_idx.remove(test_idx)
        else:
            validation = merged_data.shard(self.num_folds, fold_id)
        training_idx.remove(fold_id)
        train = None
        for f_id in training_idx:
            if not train:
                train = merged_data.shard(self.num_folds, f_id)
            else:
                train = train.concatenate(merged_data.shard(self.num_folds, f_id))

        # batchify training, validation and test sets.
        validation = validation.batch(self.batch_size, drop_remainder=False)
        train = train.batch(self.batch_size, drop_remainder=False)
        # We apply prefetch as it improved train/test/validation throughput by 30% in some real model training.
        self.train = train.prefetch(tf.data.experimental.AUTOTUNE)
        self.validation = validation.prefetch(tf.data.experimental.AUTOTUNE)
        if self.include_testset_in_kfold:
            test = test.batch(self.batch_size, drop_remainder=False)
            self.test = test.prefetch(tf.data.experimental.AUTOTUNE)
        else:
            self.test = relevance_dataset.test



class RelevanceLossBase(object):
    """
    Abstract class that defines the loss and final activation function
    used to train a RelevanceModel
    """

    def get_loss_fn(self, **kwargs):
        """
        Returns the loss function _loss_fn()

        Parameters
        ----------
        kwargs : dict
            Additional key value arguments can be passed as needed.
            For example, metadata features can be passed to compute custom losses

        Returns
        -------
        function
            Loss function that computes the loss from predicted scores
            and true labels
        """

        def _loss_fn(y_true, y_pred):
            pass

        return _loss_fn

    def get_final_activation_op(self):
        """
        Returns the final activation layer

        Returns
        -------
        function
            Final activation function that is applied on the scores
            before computing losses
        """
        raise NotImplementedError



from ml4ir.base.config.keys import ArchitectureKey
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.model.architectures.dnn import DNN


def get_architecture(model_config: dict, feature_config: FeatureConfig, file_io):
    """
    Return the architecture operation based on the model_config YAML specified
    """
    architecture_key = model_config.get("architecture_key")
    if architecture_key == ArchitectureKey.DNN:
        return DNN(model_config, feature_config, file_io).get_architecture_op()
    elif architecture_key == ArchitectureKey.LINEAR:
        # Validate the model config
        num_dense_layers = len([l for l in model_config["layers"] if l["type"] == "dense"])
        if num_dense_layers == 0:
            raise ValueError("No dense layers were specified in the ModelConfig")
        elif num_dense_layers > 1:
            raise ValueError("Linear model used with more than 1 dense layer")
        else:
            return DNN(model_config, feature_config, file_io).get_architecture_op()
    elif architecture_key == ArchitectureKey.RNN:
        raise NotImplementedError

    else:
        raise NotImplementedError



from argparse import Namespace

from ml4ir.base.config.parse_args import RelevanceArgParser
from ml4ir.applications.ranking.config.keys import ScoringTypeKey, LossTypeKey

from typing import List


class RankingArgParser(RelevanceArgParser):
    def define_args(self):
        super().define_args()

        self.add_argument(
            "--scoring_type",
            type=str,
            choices=LossTypeKey.get_all_keys(),
            default=LossTypeKey.POINTWISE,
            help="Scoring technique to use. Has to be one of the scoring types in ScoringTypeKey in "
            "applications/ranking/config/keys.py",
        )

        self.add_argument(
            "--loss_type",
            type=str,
            choices=ScoringTypeKey.get_all_keys(),
            default=ScoringTypeKey.LISTWISE,
            help="Loss technique to use. Has to be one of the loss types in LossTypeKey in "
            "applications/ranking/config/keys.py",
        )

    def set_default_args(self):
        super().set_default_args()

        self.set_defaults(
            tfrecord_type="sequence_example",
            loss_key="sigmoid_cross_entropy",
            metrics_keys=['MRR', 'ACR'],
            monitor_metric="new_MRR",
            monitor_mode="max",
            max_sequence_size=25,
            group_metrics_min_queries=25,
            output_name="ranking_score",
        )


def get_args(args: List[str]) -> Namespace:
    return RankingArgParser().parse_args(args)



import tensorflow as tf
from tensorflow import keras
from typing import List, Dict
import numpy as np
from tensorflow.keras import backend as K

from ml4ir.base.config.keys import TFRecordTypeKey
from ml4ir.base.features.feature_config import FeatureConfig


def get_predict_fn(
    model: keras.Model,
    tfrecord_type: str,
    feature_config: FeatureConfig,
    inference_signature: str = "serving_default",
    is_compiled: bool = False,
    output_name: str = "relevance_score",
    features_to_return: List = [],
    additional_features: Dict = {},
    max_sequence_size: int = 0,
):
    """
    Define a prediction function to convert input features into scores.

    Parameters
    ----------
    model : `keras.Model`
        Tensorflow keras model to be used for prediction
    tfrecord_type : {"example", "sequence_example"}
        Type of the TFRecord data we want to run prediction with the model on.
    feature_config : `FeatureConfig` object
        FeatureConfig object that defines the input features for the model
        and their respective configurations
    inference_signature : str, optional
        The name of the serving signature to be fetched from the loaded
        keras model. This will be used only if the model is not compiled into
        a Keras model.
    is_compiled : bool, optional
        Value specifying if the keras model has been compiled or not
    output_name : str, optional
        Name of the output score from the dictionary of outputs
    features_to_return : list, optional
        List of output features to fetch along with the score
    additional_features : dict, optional
        Dictionary of new feature names and corresponding functions to
        compute them. Use this to compute additional features to compute
        further metrics.
    max_sequence_size : int, optional
        Maximum size of the sequence in a TFRecord SequenceExample protobuf
        object

    Returns
    -------
    `tf.function`
        Tensorflow function that accepts features as input and returns the scores
        and other specified outputs as a dictionary

    """
    # Load the forward pass function for the tensorflow model
    if is_compiled:
        infer = model
    else:
        # If SavedModel was loaded without compilation
        infer = model.signatures[inference_signature]

    # Get features to log
    features_to_log = [f.get("node_name", f["name"]) for f in features_to_return] + [output_name]

    @tf.function
    def _flatten_records(x):
        """Collapse first two dimensions -> [batch_size, max_sequence_size]"""
        return tf.reshape(x, tf.concat([[-1], tf.shape(x)[2:]], axis=0))

    @tf.function
    def _filter_records(x, mask):
        """Filter records that were padded in each query"""
        return tf.squeeze(
            tf.gather_nd(
                x, tf.where(tf.not_equal(tf.cast(mask, tf.int64), tf.constant(0, dtype="int64"))),
            )
        )

    @tf.function
    def _predict_score(features, label):
        """Predict scores and compute additional output from input features using the input model"""
        if is_compiled:
            scores = infer(features)[output_name]
        else:
            scores = infer(**features)[output_name]

        # Set scores of padded records to 0
        if tfrecord_type == TFRecordTypeKey.SEQUENCE_EXAMPLE:
            scores = tf.where(tf.equal(features["mask"], 0), tf.constant(-np.inf), scores)

            mask = _flatten_records(features["mask"])

        predictions_dict = dict()
        for feature_name in features_to_log:
            if feature_name == feature_config.get_label(key="node_name"):
                feat_ = label
            elif feature_name == output_name:
                feat_ = scores
            else:
                if feature_name in features:
                    feat_ = features[feature_name]
                else:
                    raise KeyError("{} was not found in input training data".format(feature_name))

            predictions_dict[feature_name] = feat_

        # Process additional features
        for feature_name, feature_fn in additional_features.items():
            predictions_dict[feature_name] = feature_fn(features, label, scores)

        # Explode context features to each record for logging
        # NOTE: This assumes that the record dimension is on axis 1, like previously
        for feature_name in predictions_dict.keys():
            feat_ = predictions_dict[feature_name]
            if tfrecord_type == TFRecordTypeKey.SEQUENCE_EXAMPLE:
                feat_ = tf.cond(
                    tf.equal(tf.shape(feat_)[1], tf.constant(1)),
                    true_fn=lambda: K.repeat_elements(feat_, rep=max_sequence_size, axis=1),
                    false_fn=lambda: feat_,
                )

                # Collapse from one query per data point to one record per data point
                # and remove padded dummy records
                feat_ = _filter_records(_flatten_records(feat_), mask)

            predictions_dict[feature_name] = tf.squeeze(feat_)

        return predictions_dict

    return _predict_score



import glob
import os
from typing import Optional
from logging import Logger
import tensorflow as tf

from ml4ir.base.config.keys import DataFormatKey, DataSplitKey
from ml4ir.base.data import csv_reader
from ml4ir.base.data import tfrecord_reader
from ml4ir.base.data import ranklib_reader
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.io.file_io import FileIO


class RelevanceDataset:
    """class to create/load TFRecordDataset for train, validation and test"""

    def __init__(
        self,
        data_dir: str,
        data_format: str,
        feature_config: FeatureConfig,
        tfrecord_type: str,
        file_io: FileIO,
        max_sequence_size: int = 0,
        batch_size: int = 128,
        preprocessing_keys_to_fns: dict = {},
        train_pcent_split: float = 0.8,
        val_pcent_split: float = -1,
        test_pcent_split: float = -1,
        use_part_files: bool = False,
        parse_tfrecord: bool = True,
        logger: Optional[Logger] = None,
        keep_additional_info: int = 0,
        non_zero_features_only: int = 0,
    ):
        """
        Constructor method to instantiate a RelevanceDataset object
        Loads and creates the TFRecordDataset for train, validation and test splits

        Parameters
        ----------
        data_dir : str
            path to the directory containing train, validation and test data
        data_format : {"tfrecord", "csv", "libsvm"}
            type of data files to be converted into TFRecords and loaded as a TFRecordDataset
        feature_config : `FeatureConfig` object
            FeatureConfig object that defines the features to be loaded in the dataset
            and the preprocessing functions to be applied to each of them
        tfrecord_type : {"example", "sequence_example"}
            Type of the TFRecord protobuf message to be used for TFRecordDataset
        file_io : `FileIO` object
            file I/O handler objects for reading and writing data
        max_sequence_size : int, optional
            maximum number of sequence to be used with a single SequenceExample proto message
            The data will be appropriately padded or clipped to fit the max value specified
        batch_size : int, optional
            size of each data batch
        preprocessing_keys_to_fns : dict of (str, function), optional
            dictionary of function names mapped to function definitions
            that can now be used for preprocessing while loading the
            TFRecordDataset to create the RelevanceDataset object
        train_pcent_split : float, optional
            ratio of overall data to be used as training set
        val_pcent_split : float, optional
            ratio of overall data to be used as validation set
        test_pcent_split : float, optional
            ratio of overall data to be used as test set
        use_part_files : bool, optional
            load dataset from part files checked using "part-" prefix
        parse_tfrecord : bool, optional
            parse the TFRecord string from the dataset;
            returns strings as is otherwise
        logger : `Logger`, optional
            logging handler for status messages

        Notes
        -----
        * Currently supports CSV, TFRecord and Libsvm data formats
        * Does not support automatically splitting train, validation and test
        * `data_dir` should contain `train`, `validation` and `test` directories with files within them
        """
        self.feature_config = feature_config
        self.max_sequence_size = max_sequence_size
        self.logger = logger
        self.data_dir = data_dir
        self.data_format: str = data_format
        self.tfrecord_type = tfrecord_type
        self.batch_size: int = batch_size
        self.preprocessing_keys_to_fns = preprocessing_keys_to_fns
        self.file_io = file_io

        self.train_pcent_split: float = train_pcent_split
        self.val_pcent_split: float = val_pcent_split
        self.test_pcent_split: float = test_pcent_split
        self.use_part_files: bool = use_part_files

        self.keep_additional_info = keep_additional_info
        self.non_zero_features_only = non_zero_features_only

        self.train: Optional[tf.data.TFRecordDataset] = None
        self.validation: Optional[tf.data.TFRecordDataset] = None
        self.test: Optional[tf.data.TFRecordDataset] = None
        self.create_dataset(parse_tfrecord)

    def create_dataset(self, parse_tfrecord=True):
        """
        Loads and creates train, validation and test datasets

        Parameters
        ----------
        parse_tfrecord : bool
            parse the TFRecord string from the dataset;
            returns strings as is otherwise
        """
        to_split = len(glob.glob(os.path.join(self.data_dir, DataSplitKey.TEST))) == 0

        if self.data_format == DataFormatKey.CSV:
            data_reader = csv_reader
        elif self.data_format == DataFormatKey.TFRECORD:
            data_reader = tfrecord_reader
        elif self.data_format == DataFormatKey.RANKLIB:
            data_reader = ranklib_reader
        else:
            raise NotImplementedError(
                "Unsupported data format: {}. We currenty support {} and {}.".format(
                    self.data_format, DataFormatKey.CSV, DataFormatKey.TFRECORD
                )
            )

        if to_split:
            """
            If the data is stored as
            data_dir
            │
            ├── data_file
            ├── data_file
            ├── ...
            └── data_file
            """
            raise NotImplementedError

        else:
            """
            If the data is stored as
            data_dir
            │
            ├── train
            │   ├── data_file
            │   ├── data_file
            │   ├── ...
            │   └── data_file
            ├── validation
            │   ├── data_file
            │   ├── data_file
            │   ├── ...
            │   └── data_file
            └── test
                ├── data_file
                ├── data_file
                ├── ...
                └── data_file

            We also apply prefetch(tf.data.experimental.AUTOTUNE)
            as it improved train/test/validation throughput
            by 30% in some real model training.
            """
            self.train = data_reader.read(
                data_dir=os.path.join(self.data_dir, DataSplitKey.TRAIN),
                feature_config=self.feature_config,
                tfrecord_type=self.tfrecord_type,
                tfrecord_dir=os.path.join(self.data_dir, "tfrecord", DataSplitKey.TRAIN),
                max_sequence_size=self.max_sequence_size,
                batch_size=self.batch_size,
                preprocessing_keys_to_fns=self.preprocessing_keys_to_fns,
                use_part_files=self.use_part_files,
                parse_tfrecord=parse_tfrecord,
                file_io=self.file_io,
                logger=self.logger,
                keep_additional_info=self.keep_additional_info,
                non_zero_features_only=self.non_zero_features_only,
            )
            self.validation = data_reader.read(
                data_dir=os.path.join(self.data_dir, DataSplitKey.VALIDATION),
                feature_config=self.feature_config,
                tfrecord_type=self.tfrecord_type,
                tfrecord_dir=os.path.join(self.data_dir, "tfrecord", DataSplitKey.VALIDATION),
                max_sequence_size=self.max_sequence_size,
                batch_size=self.batch_size,
                preprocessing_keys_to_fns=self.preprocessing_keys_to_fns,
                use_part_files=self.use_part_files,
                parse_tfrecord=parse_tfrecord,
                file_io=self.file_io,
                logger=self.logger,
                keep_additional_info=self.keep_additional_info,
                non_zero_features_only=self.non_zero_features_only,
            )
            self.test = data_reader.read(
                data_dir=os.path.join(self.data_dir, DataSplitKey.TEST),
                feature_config=self.feature_config,
                tfrecord_type=self.tfrecord_type,
                tfrecord_dir=os.path.join(self.data_dir, "tfrecord", DataSplitKey.TEST),
                max_sequence_size=self.max_sequence_size,
                batch_size=self.batch_size,
                preprocessing_keys_to_fns=self.preprocessing_keys_to_fns,
                use_part_files=self.use_part_files,
                parse_tfrecord=parse_tfrecord,
                file_io=self.file_io,
                logger=self.logger,
                keep_additional_info=self.keep_additional_info,
                non_zero_features_only=self.non_zero_features_only,
            )

    def balance_classes(self):
        """
        Balance class labels in the train dataset
        """
        raise NotImplementedError

    def train_val_test_split(self):
        """
        Split the dataset into train, validation and test
        """
        raise NotImplementedError
